from dataclasses import dataclass
from dbt.adapters.reference_keys import _ReferenceKey
from dbt import ui
from dbt.helper_types import Lazy
from dbt.events.base_types import (
    Event,
    NoFile,
    DebugLevel,
    InfoLevel,
    WarnLevel,
    ErrorLevel,
    ShowException,
    NodeInfo,
    Cache,
)
from dbt.events.format import format_fancy_output_line, pluralize
from dbt.events.serialization import EventSerialization
from dbt.node_types import NodeType
from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar


# The classes in this file represent the data necessary to describe a
# particular event to both human readable logs, and machine reliable
# event streams. classes extend superclasses that indicate what
# destinations they are intended for, which mypy uses to enforce
# that the necessary methods are defined.


# Type representing Event and all subclasses of Event
T_Event = TypeVar("T_Event", bound=Event)


# Event codes have prefixes which follow this table
#
# | Code |     Description     |
# |:----:|:-------------------:|
# | A    | Pre-project loading |
# | E    | DB adapter          |
# | I    | Project parsing     |
# | M    | Deps generation     |
# | Q    | Node processing     |
# | W    | Node testing        |
# | Y    | Post processing     |
# | Z    | Misc                |
# | T    | Test only           |
#
# The basic idea is that event codes roughly translate to the natural order of running a dbt task

# TODO: remove ingore when this is fixed:
# https://github.com/python/mypy/issues/5374


@dataclass  # type: ignore
class AdapterEventBase(EventSerialization, Event):
    name: str
    base_msg: str
    args: Tuple[Any, ...]

    # instead of having this inherit from one of the level classes
    def level_tag(self) -> str:
        raise Exception("level_tag should never be called on AdapterEventBase")

    def message(self) -> str:
        # this class shouldn't be createable, but we can't make it an ABC because of a mypy bug
        if type(self).__name__ == "AdapterEventBase":
            raise Exception(
                "attempted to create a message for AdapterEventBase which cannot be created"
            )

        # only apply formatting if there are arguments to format.
        # avoids issues like "dict: {k: v}".format() which results in `KeyError 'k'`
        msg = self.base_msg if len(self.args) == 0 else self.base_msg.format(*self.args)
        return f"{self.name} adapter: {msg}"


@dataclass
class AdapterEventDebug(DebugLevel, AdapterEventBase, ShowException):
    code: str = "E001"


@dataclass
class AdapterEventInfo(InfoLevel, AdapterEventBase, ShowException):
    code: str = "E002"


@dataclass
class AdapterEventWarning(WarnLevel, AdapterEventBase, ShowException):
    code: str = "E003"


@dataclass
class AdapterEventError(ErrorLevel, AdapterEventBase, ShowException):
    code: str = "E004"


@dataclass
class MainKeyboardInterrupt(InfoLevel):
    code: str = "Z001"

    def message(self) -> str:
        return "ctrl-c"


@dataclass
class MainEncounteredError(ErrorLevel):
    e: BaseException
    code: str = "Z002"

    def message(self) -> str:
        return f"Encountered an error:\n{self.e}"


@dataclass
class MainStackTrace(ErrorLevel):
    stack_trace: str
    code: str = "Z003"

    def message(self) -> str:
        return self.stack_trace


@dataclass
class MainReportVersion(InfoLevel):
    v: str  # could be VersionSpecifier instead if we resolved some circular imports
    code: str = "A001"

    def message(self):
        return f"Running with dbt{self.v}"


@dataclass
class MainReportArgs(DebugLevel):
    args: Dict[str, str]
    code: str = "A002"

    def message(self):
        return f"running dbt with arguments {str(self.args)}"


@dataclass
class MainTrackingUserState(DebugLevel):
    user_state: str
    code: str = "A003"

    def message(self):
        return f"Tracking: {self.user_state}"


@dataclass
class ParsingStart(InfoLevel):
    code: str = "I001"

    def message(self) -> str:
        return "Start parsing."


@dataclass
class ParsingCompiling(InfoLevel):
    code: str = "I002"

    def message(self) -> str:
        return "Compiling."


@dataclass
class ParsingWritingManifest(InfoLevel):
    code: str = "I003"

    def message(self) -> str:
        return "Writing manifest."


@dataclass
class ParsingDone(InfoLevel):
    code: str = "I004"

    def message(self) -> str:
        return "Done."


@dataclass
class ManifestDependenciesLoaded(InfoLevel):
    code: str = "I005"

    def message(self) -> str:
        return "Dependencies loaded"


@dataclass
class ManifestLoaderCreated(InfoLevel):
    code: str = "I006"

    def message(self) -> str:
        return "ManifestLoader created"


@dataclass
class ManifestLoaded(InfoLevel):
    code: str = "I007"

    def message(self) -> str:
        return "Manifest loaded"


@dataclass
class ManifestChecked(InfoLevel):
    code: str = "I008"

    def message(self) -> str:
        return "Manifest checked"


@dataclass
class ManifestFlatGraphBuilt(InfoLevel):
    code: str = "I009"

    def message(self) -> str:
        return "Flat graph built"


@dataclass
class ReportPerformancePath(InfoLevel):
    path: str
    code: str = "I010"

    def message(self) -> str:
        return f"Performance info: {self.path}"


@dataclass
class GitSparseCheckoutSubdirectory(DebugLevel):
    subdir: str
    code: str = "M001"

    def message(self) -> str:
        return f"  Subdirectory specified: {self.subdir}, using sparse checkout."


@dataclass
class GitProgressCheckoutRevision(DebugLevel):
    revision: str
    code: str = "M002"

    def message(self) -> str:
        return f"  Checking out revision {self.revision}."


@dataclass
class GitProgressUpdatingExistingDependency(DebugLevel):
    dir: str
    code: str = "M003"

    def message(self) -> str:
        return f"Updating existing dependency {self.dir}."


@dataclass
class GitProgressPullingNewDependency(DebugLevel):
    dir: str
    code: str = "M004"

    def message(self) -> str:
        return f"Pulling new dependency {self.dir}."


@dataclass
class GitNothingToDo(DebugLevel):
    sha: str
    code: str = "M005"

    def message(self) -> str:
        return f"Already at {self.sha}, nothing to do."


@dataclass
class GitProgressUpdatedCheckoutRange(DebugLevel):
    start_sha: str
    end_sha: str
    code: str = "M006"

    def message(self) -> str:
        return f"  Updated checkout from {self.start_sha} to {self.end_sha}."


@dataclass
class GitProgressCheckedOutAt(DebugLevel):
    end_sha: str
    code: str = "M007"

    def message(self) -> str:
        return f"  Checked out at {self.end_sha}."


@dataclass
class RegistryIndexProgressMakingGETRequest(DebugLevel):
    url: str
    code: str = "M022"

    def message(self) -> str:
        return f"Making package index registry request: GET {self.url}"


@dataclass
class RegistryIndexProgressGETResponse(DebugLevel):
    url: str
    resp_code: int
    code: str = "M023"

    def message(self) -> str:
        return f"Response from registry index: GET {self.url} {self.resp_code}"


@dataclass
class RegistryProgressMakingGETRequest(DebugLevel):
    url: str
    code: str = "M008"

    def message(self) -> str:
        return f"Making package registry request: GET {self.url}"


@dataclass
class RegistryProgressGETResponse(DebugLevel):
    url: str
    resp_code: int
    code: str = "M009"

    def message(self) -> str:
        return f"Response from registry: GET {self.url} {self.resp_code}"


@dataclass
class RegistryResponseUnexpectedType(DebugLevel):
    response: str
    code: str = "M024"

    def message(self) -> str:
        return f"Response was None: {self.response}"


@dataclass
class RegistryResponseMissingTopKeys(DebugLevel):
    response: str
    code: str = "M025"

    def message(self) -> str:
        # expected/actual keys logged in exception
        return f"Response missing top level keys: {self.response}"


@dataclass
class RegistryResponseMissingNestedKeys(DebugLevel):
    response: str
    code: str = "M026"

    def message(self) -> str:
        # expected/actual keys logged in exception
        return f"Response missing nested keys: {self.response}"


@dataclass
class RegistryResponseExtraNestedKeys(DebugLevel):
    response: str
    code: str = "M027"

    def message(self) -> str:
        # expected/actual keys logged in exception
        return f"Response contained inconsistent keys: {self.response}"


# TODO this was actually `logger.exception(...)` not `logger.error(...)`
@dataclass
class SystemErrorRetrievingModTime(ErrorLevel):
    path: str
    code: str = "Z004"

    def message(self) -> str:
        return f"Error retrieving modification time for file {self.path}"


@dataclass
class SystemCouldNotWrite(DebugLevel):
    path: str
    reason: str
    exc: Exception
    code: str = "Z005"

    def message(self) -> str:
        return (
            f"Could not write to path {self.path}({len(self.path)} characters): "
            f"{self.reason}\nexception: {self.exc}"
        )


@dataclass
class SystemExecutingCmd(DebugLevel):
    cmd: List[str]
    code: str = "Z006"

    def message(self) -> str:
        return f'Executing "{" ".join(self.cmd)}"'


@dataclass
class SystemStdOutMsg(DebugLevel):
    bmsg: bytes
    code: str = "Z007"

    def message(self) -> str:
        return f'STDOUT: "{str(self.bmsg)}"'


@dataclass
class SystemStdErrMsg(DebugLevel):
    bmsg: bytes
    code: str = "Z008"

    def message(self) -> str:
        return f'STDERR: "{str(self.bmsg)}"'


@dataclass
class SystemReportReturnCode(DebugLevel):
    returncode: int
    code: str = "Z009"

    def message(self) -> str:
        return f"command return code={self.returncode}"


@dataclass
class SelectorReportInvalidSelector(InfoLevel):
    valid_selectors: str
    spec_method: str
    raw_spec: str
    code: str = "M010"

    def message(self) -> str:
        return (
            f"The '{self.spec_method}' selector specified in {self.raw_spec} is "
            f"invalid. Must be one of [{self.valid_selectors}]"
        )


@dataclass
class MacroEventInfo(InfoLevel):
    msg: str
    code: str = "M011"

    def message(self) -> str:
        return self.msg


@dataclass
class MacroEventDebug(DebugLevel):
    msg: str
    code: str = "M012"

    def message(self) -> str:
        return self.msg


@dataclass
class NewConnection(DebugLevel):
    conn_type: str
    conn_name: str
    code: str = "E005"

    def message(self) -> str:
        return f'Acquiring new {self.conn_type} connection "{self.conn_name}"'


@dataclass
class ConnectionReused(DebugLevel):
    conn_name: str
    code: str = "E006"

    def message(self) -> str:
        return f"Re-using an available connection from the pool (formerly {self.conn_name})"


@dataclass
class ConnectionLeftOpen(DebugLevel):
    conn_name: Optional[str]
    code: str = "E007"

    def message(self) -> str:
        return f"Connection '{self.conn_name}' was left open."


@dataclass
class ConnectionClosed(DebugLevel):
    conn_name: Optional[str]
    code: str = "E008"

    def message(self) -> str:
        return f"Connection '{self.conn_name}' was properly closed."


@dataclass
class RollbackFailed(ShowException, DebugLevel):
    conn_name: Optional[str]
    code: str = "E009"

    def message(self) -> str:
        return f"Failed to rollback '{self.conn_name}'"


# TODO: can we combine this with ConnectionClosed?
@dataclass
class ConnectionClosed2(DebugLevel):
    conn_name: Optional[str]
    code: str = "E010"

    def message(self) -> str:
        return f"On {self.conn_name}: Close"


# TODO: can we combine this with ConnectionLeftOpen?
@dataclass
class ConnectionLeftOpen2(DebugLevel):
    conn_name: Optional[str]
    code: str = "E011"

    def message(self) -> str:
        return f"On {self.conn_name}: No close available on handle"


@dataclass
class Rollback(DebugLevel):
    conn_name: Optional[str]
    code: str = "E012"

    def message(self) -> str:
        return f"On {self.conn_name}: ROLLBACK"


@dataclass
class CacheMiss(DebugLevel):
    conn_name: str
    database: Optional[str]
    schema: str
    code: str = "E013"

    def message(self) -> str:
        return (
            f'On "{self.conn_name}": cache miss for schema '
            '"{self.database}.{self.schema}", this is inefficient'
        )


@dataclass
class ListRelations(DebugLevel):
    database: Optional[str]
    schema: str
    relations: List[_ReferenceKey]
    code: str = "E014"

    def message(self) -> str:
        return f"with database={self.database}, schema={self.schema}, relations={self.relations}"


@dataclass
class ConnectionUsed(DebugLevel):
    conn_type: str
    conn_name: Optional[str]
    code: str = "E015"

    def message(self) -> str:
        return f'Using {self.conn_type} connection "{self.conn_name}"'


@dataclass
class SQLQuery(DebugLevel):
    conn_name: Optional[str]
    sql: str
    code: str = "E016"

    def message(self) -> str:
        return f"On {self.conn_name}: {self.sql}"


@dataclass
class SQLQueryStatus(DebugLevel):
    status: str
    elapsed: Optional[float]
    code: str = "E017"

    def message(self) -> str:
        return f"SQL status: {self.status} in {self.elapsed} seconds"


@dataclass
class SQLCommit(DebugLevel):
    conn_name: str
    code: str = "E018"

    def message(self) -> str:
        return f"On {self.conn_name}: COMMIT"


@dataclass
class ColTypeChange(DebugLevel):
    orig_type: str
    new_type: str
    table: _ReferenceKey
    code: str = "E019"

    def message(self) -> str:
        return f"Changing col type from {self.orig_type} to {self.new_type} in table {self.table}"


@dataclass
class SchemaCreation(DebugLevel):
    relation: _ReferenceKey
    code: str = "E020"

    def message(self) -> str:
        return f'Creating schema "{self.relation}"'


@dataclass
class SchemaDrop(DebugLevel):
    relation: _ReferenceKey
    code: str = "E021"

    def message(self) -> str:
        return f'Dropping schema "{self.relation}".'


# TODO pretty sure this is only ever called in dead code
# see: core/dbt/adapters/cache.py _add_link vs add_link
@dataclass
class UncachedRelation(DebugLevel, Cache):
    dep_key: _ReferenceKey
    ref_key: _ReferenceKey
    code: str = "E022"

    def message(self) -> str:
        return (
            f"{self.dep_key} references {str(self.ref_key)} "
            "but {self.ref_key.database}.{self.ref_key.schema}"
            "is not in the cache, skipping assumed external relation"
        )


@dataclass
class AddLink(DebugLevel, Cache):
    dep_key: _ReferenceKey
    ref_key: _ReferenceKey
    code: str = "E023"

    def message(self) -> str:
        return f"adding link, {self.dep_key} references {self.ref_key}"


@dataclass
class AddRelation(DebugLevel, Cache):
    relation: _ReferenceKey
    code: str = "E024"

    def message(self) -> str:
        return f"Adding relation: {str(self.relation)}"


@dataclass
class DropMissingRelation(DebugLevel, Cache):
    relation: _ReferenceKey
    code: str = "E025"

    def message(self) -> str:
        return f"dropped a nonexistent relationship: {str(self.relation)}"


@dataclass
class DropCascade(DebugLevel, Cache):
    dropped: _ReferenceKey
    consequences: Set[_ReferenceKey]
    code: str = "E026"

    def message(self) -> str:
        return f"drop {self.dropped} is cascading to {self.consequences}"


@dataclass
class DropRelation(DebugLevel, Cache):
    dropped: _ReferenceKey
    code: str = "E027"

    def message(self) -> str:
        return f"Dropping relation: {self.dropped}"


@dataclass
class UpdateReference(DebugLevel, Cache):
    old_key: _ReferenceKey
    new_key: _ReferenceKey
    cached_key: _ReferenceKey
    code: str = "E028"

    def message(self) -> str:
        return (
            f"updated reference from {self.old_key} -> {self.cached_key} to "
            "{self.new_key} -> {self.cached_key}"
        )


@dataclass
class TemporaryRelation(DebugLevel, Cache):
    key: _ReferenceKey
    code: str = "E029"

    def message(self) -> str:
        return f"old key {self.key} not found in self.relations, assuming temporary"


@dataclass
class RenameSchema(DebugLevel, Cache):
    old_key: _ReferenceKey
    new_key: _ReferenceKey
    code: str = "E030"

    def message(self) -> str:
        return f"Renaming relation {self.old_key} to {self.new_key}"


@dataclass
class DumpBeforeAddGraph(DebugLevel, Cache):
    dump: Lazy[Dict[str, List[str]]]
    code: str = "E031"

    def message(self) -> str:
        return f"before adding : {self.dump.force()}"


@dataclass
class DumpAfterAddGraph(DebugLevel, Cache):
    dump: Lazy[Dict[str, List[str]]]
    code: str = "E032"

    def message(self) -> str:
        return f"after adding: {self.dump.force()}"


@dataclass
class DumpBeforeRenameSchema(DebugLevel, Cache):
    dump: Lazy[Dict[str, List[str]]]
    code: str = "E033"

    def message(self) -> str:
        return f"before rename: {self.dump.force()}"


@dataclass
class DumpAfterRenameSchema(DebugLevel, Cache):
    dump: Lazy[Dict[str, List[str]]]
    code: str = "E034"

    def message(self) -> str:
        return f"after rename: {self.dump.force()}"


@dataclass
class AdapterImportError(InfoLevel):
    exc: Exception
    code: str = "E035"

    def message(self) -> str:
        return f"Error importing adapter: {self.exc}"


@dataclass
class PluginLoadError(ShowException, DebugLevel):
    code: str = "E036"

    def message(self):
        pass


@dataclass
class NewConnectionOpening(DebugLevel):
    connection_state: str
    code: str = "E037"

    def message(self) -> str:
        return f"Opening a new connection, currently in state {self.connection_state}"


@dataclass
class TimingInfoCollected(DebugLevel):
    code: str = "Z010"

    def message(self) -> str:
        return "finished collecting timing info"


@dataclass
class MergedFromState(DebugLevel):
    nbr_merged: int
    sample: List
    code: str = "A004"

    def message(self) -> str:
        return f"Merged {self.nbr_merged} items from state (sample: {self.sample})"


@dataclass
class MissingProfileTarget(InfoLevel):
    profile_name: str
    target_name: str
    code: str = "A005"

    def message(self) -> str:
        return f"target not specified in profile '{self.profile_name}', using '{self.target_name}'"


@dataclass
class InvalidVarsYAML(ErrorLevel):
    code: str = "A008"

    def message(self) -> str:
        return "The YAML provided in the --vars argument is not valid."


@dataclass
class GenericTestFileParse(DebugLevel):
    path: str
    code: str = "I011"

    def message(self) -> str:
        return f"Parsing {self.path}"


@dataclass
class MacroFileParse(DebugLevel):
    path: str
    code: str = "I012"

    def message(self) -> str:
        return f"Parsing {self.path}"


@dataclass
class PartialParsingFullReparseBecauseOfError(InfoLevel):
    code: str = "I013"

    def message(self) -> str:
        return "Partial parsing enabled but an error occurred. Switching to a full re-parse."


@dataclass
class PartialParsingExceptionFile(DebugLevel):
    file: str
    code: str = "I014"

    def message(self) -> str:
        return f"Partial parsing exception processing file {self.file}"


@dataclass
class PartialParsingFile(DebugLevel):
    file_dict: Dict
    code: str = "I015"

    def message(self) -> str:
        return f"PP file: {self.file_dict}"


@dataclass
class PartialParsingException(DebugLevel):
    exc_info: Dict
    code: str = "I016"

    def message(self) -> str:
        return f"PP exception info: {self.exc_info}"


@dataclass
class PartialParsingSkipParsing(DebugLevel):
    code: str = "I017"

    def message(self) -> str:
        return "Partial parsing enabled, no changes found, skipping parsing"


@dataclass
class PartialParsingMacroChangeStartFullParse(InfoLevel):
    code: str = "I018"

    def message(self) -> str:
        return "Change detected to override macro used during parsing. Starting full parse."


@dataclass
class PartialParsingProjectEnvVarsChanged(InfoLevel):
    code: str = "I019"

    def message(self) -> str:
        return "Unable to do partial parsing because env vars used in dbt_project.yml have changed"


@dataclass
class PartialParsingProfileEnvVarsChanged(InfoLevel):
    code: str = "I020"

    def message(self) -> str:
        return "Unable to do partial parsing because env vars used in profiles.yml have changed"


@dataclass
class PartialParsingDeletedMetric(DebugLevel):
    id: str
    code: str = "I021"

    def message(self) -> str:
        return f"Partial parsing: deleted metric {self.id}"


@dataclass
class ManifestWrongMetadataVersion(DebugLevel):
    version: str
    code: str = "I022"

    def message(self) -> str:
        return (
            "Manifest metadata did not contain correct version. "
            f"Contained '{self.version}' instead."
        )


@dataclass
class PartialParsingVersionMismatch(InfoLevel):
    saved_version: str
    current_version: str
    code: str = "I023"

    def message(self) -> str:
        return (
            "Unable to do partial parsing because of a dbt version mismatch. "
            f"Saved manifest version: {self.saved_version}. "
            f"Current version: {self.current_version}."
        )


@dataclass
class PartialParsingFailedBecauseConfigChange(InfoLevel):
    code: str = "I024"

    def message(self) -> str:
        return (
            "Unable to do partial parsing because config vars, "
            "config profile, or config target have changed"
        )


@dataclass
class PartialParsingFailedBecauseProfileChange(InfoLevel):
    code: str = "I025"

    def message(self) -> str:
        return "Unable to do partial parsing because profile has changed"


@dataclass
class PartialParsingFailedBecauseNewProjectDependency(InfoLevel):
    code: str = "I026"

    def message(self) -> str:
        return "Unable to do partial parsing because a project dependency has been added"


@dataclass
class PartialParsingFailedBecauseHashChanged(InfoLevel):
    code: str = "I027"

    def message(self) -> str:
        return "Unable to do partial parsing because a project config has changed"


@dataclass
class PartialParsingNotEnabled(DebugLevel):
    code: str = "I028"

    def message(self) -> str:
        return "Partial parsing not enabled"


@dataclass
class ParsedFileLoadFailed(ShowException, DebugLevel):
    path: str
    exc: Exception
    code: str = "I029"

    def message(self) -> str:
        return f"Failed to load parsed file from disk at {self.path}: {self.exc}"


@dataclass
class PartialParseSaveFileNotFound(InfoLevel):
    code: str = "I030"

    def message(self) -> str:
        return "Partial parse save file not found. Starting full parse."


@dataclass
class StaticParserCausedJinjaRendering(DebugLevel):
    path: str
    code: str = "I031"

    def message(self) -> str:
        return f"1605: jinja rendering because of STATIC_PARSER flag. file: {self.path}"


# TODO: Experimental/static parser uses these for testing and some may be a good use case for
#       the `TestLevel` logger once we implement it.  Some will probably stay `DebugLevel`.
@dataclass
class UsingExperimentalParser(DebugLevel):
    path: str
    code: str = "I032"

    def message(self) -> str:
        return f"1610: conducting experimental parser sample on {self.path}"


@dataclass
class SampleFullJinjaRendering(DebugLevel):
    path: str
    code: str = "I033"

    def message(self) -> str:
        return f"1611: conducting full jinja rendering sample on {self.path}"


@dataclass
class StaticParserFallbackJinjaRendering(DebugLevel):
    path: str
    code: str = "I034"

    def message(self) -> str:
        return f"1602: parser fallback to jinja rendering on {self.path}"


@dataclass
class StaticParsingMacroOverrideDetected(DebugLevel):
    path: str
    code: str = "I035"

    def message(self) -> str:
        return f"1601: detected macro override of ref/source/config in the scope of {self.path}"


@dataclass
class StaticParserSuccess(DebugLevel):
    path: str
    code: str = "I036"

    def message(self) -> str:
        return f"1699: static parser successfully parsed {self.path}"


@dataclass
class StaticParserFailure(DebugLevel):
    path: str
    code: str = "I037"

    def message(self) -> str:
        return f"1603: static parser failed on {self.path}"


@dataclass
class ExperimentalParserSuccess(DebugLevel):
    path: str
    code: str = "I038"

    def message(self) -> str:
        return f"1698: experimental parser successfully parsed {self.path}"


@dataclass
class ExperimentalParserFailure(DebugLevel):
    path: str
    code: str = "I039"

    def message(self) -> str:
        return f"1604: experimental parser failed on {self.path}"


@dataclass
class PartialParsingEnabled(DebugLevel):
    deleted: int
    added: int
    changed: int
    code: str = "I040"

    def message(self) -> str:
        return (
            f"Partial parsing enabled: "
            f"{self.deleted} files deleted, "
            f"{self.added} files added, "
            f"{self.changed} files changed."
        )


@dataclass
class PartialParsingAddedFile(DebugLevel):
    file_id: str
    code: str = "I041"

    def message(self) -> str:
        return f"Partial parsing: added file: {self.file_id}"


@dataclass
class PartialParsingDeletedFile(DebugLevel):
    file_id: str
    code: str = "I042"

    def message(self) -> str:
        return f"Partial parsing: deleted file: {self.file_id}"


@dataclass
class PartialParsingUpdatedFile(DebugLevel):
    file_id: str
    code: str = "I043"

    def message(self) -> str:
        return f"Partial parsing: updated file: {self.file_id}"


@dataclass
class PartialParsingNodeMissingInSourceFile(DebugLevel):
    source_file: str
    code: str = "I044"

    def message(self) -> str:
        return f"Partial parsing: node not found for source_file {self.source_file}"


@dataclass
class PartialParsingMissingNodes(DebugLevel):
    file_id: str
    code: str = "I045"

    def message(self) -> str:
        return f"No nodes found for source file {self.file_id}"


@dataclass
class PartialParsingChildMapMissingUniqueID(DebugLevel):
    unique_id: str
    code: str = "I046"

    def message(self) -> str:
        return f"Partial parsing: {self.unique_id} not found in child_map"


@dataclass
class PartialParsingUpdateSchemaFile(DebugLevel):
    file_id: str
    code: str = "I047"

    def message(self) -> str:
        return f"Partial parsing: update schema file: {self.file_id}"


@dataclass
class PartialParsingDeletedSource(DebugLevel):
    unique_id: str
    code: str = "I048"

    def message(self) -> str:
        return f"Partial parsing: deleted source {self.unique_id}"


@dataclass
class PartialParsingDeletedExposure(DebugLevel):
    unique_id: str
    code: str = "I049"

    def message(self) -> str:
        return f"Partial parsing: deleted exposure {self.unique_id}"


@dataclass
class InvalidDisabledSourceInTestNode(WarnLevel):
    msg: str
    code: str = "I050"

    def message(self) -> str:
        return ui.warning_tag(self.msg)


@dataclass
class InvalidRefInTestNode(DebugLevel):
    msg: str
    code: str = "I051"

    def message(self) -> str:
        return self.msg


@dataclass
class RunningOperationCaughtError(ErrorLevel):
    exc: Exception
    code: str = "Q001"

    def message(self) -> str:
        return f"Encountered an error while running operation: {self.exc}"


@dataclass
class RunningOperationUncaughtError(ErrorLevel):
    exc: Exception
    code: str = "FF01"

    def message(self) -> str:
        return f"Encountered an error while running operation: {self.exc}"


@dataclass
class DbtProjectError(ErrorLevel):
    code: str = "A009"

    def message(self) -> str:
        return "Encountered an error while reading the project:"


@dataclass
class DbtProjectErrorException(ErrorLevel):
    exc: Exception
    code: str = "A010"

    def message(self) -> str:
        return f"  ERROR: {str(self.exc)}"


@dataclass
class DbtProfileError(ErrorLevel):
    code: str = "A011"

    def message(self) -> str:
        return "Encountered an error while reading profiles:"


@dataclass
class DbtProfileErrorException(ErrorLevel):
    exc: Exception
    code: str = "A012"

    def message(self) -> str:
        return f"  ERROR: {str(self.exc)}"


@dataclass
class ProfileListTitle(InfoLevel):
    code: str = "A013"

    def message(self) -> str:
        return "Defined profiles:"


@dataclass
class ListSingleProfile(InfoLevel):
    profile: str
    code: str = "A014"

    def message(self) -> str:
        return f" - {self.profile}"


@dataclass
class NoDefinedProfiles(InfoLevel):
    code: str = "A015"

    def message(self) -> str:
        return "There are no profiles defined in your profiles.yml file"


@dataclass
class ProfileHelpMessage(InfoLevel):
    code: str = "A016"

    def message(self) -> str:
        PROFILES_HELP_MESSAGE = """
For more information on configuring profiles, please consult the dbt docs:

https://docs.getdbt.com/docs/configure-your-profile
"""
        return PROFILES_HELP_MESSAGE


@dataclass
class CatchableExceptionOnRun(ShowException, DebugLevel):
    exc: Exception
    code: str = "W002"

    def message(self) -> str:
        return str(self.exc)


@dataclass
class InternalExceptionOnRun(DebugLevel):
    build_path: str
    exc: Exception
    code: str = "W003"

    def message(self) -> str:
        prefix = "Internal error executing {}".format(self.build_path)

        INTERNAL_ERROR_STRING = """This is an error in dbt. Please try again. If \
the error persists, open an issue at https://github.com/dbt-labs/dbt-core
""".strip()

        return "{prefix}\n{error}\n\n{note}".format(
            prefix=ui.red(prefix), error=str(self.exc).strip(), note=INTERNAL_ERROR_STRING
        )


# This prints the stack trace at the debug level while allowing just the nice exception message
# at the error level - or whatever other level chosen.  Used in multiple places.
@dataclass
class PrintDebugStackTrace(ShowException, DebugLevel):
    code: str = "Z011"

    def message(self) -> str:
        return ""


@dataclass
class GenericExceptionOnRun(ErrorLevel):
    build_path: Optional[str]
    unique_id: str
    exc: Exception
    code: str = "W004"

    def message(self) -> str:
        node_description = self.build_path
        if node_description is None:
            node_description = self.unique_id
        prefix = "Unhandled error while executing {}".format(node_description)
        return "{prefix}\n{error}".format(prefix=ui.red(prefix), error=str(self.exc).strip())


@dataclass
class NodeConnectionReleaseError(ShowException, DebugLevel):
    node_name: str
    exc: Exception
    code: str = "W005"

    def message(self) -> str:
        return "Error releasing connection for node {}: {!s}".format(self.node_name, self.exc)


# We don't write "clean" events to the log, because the clean command
# may have removed the log directory.
@dataclass
class CheckCleanPath(InfoLevel, NoFile):
    path: str
    code: str = "Z012"

    def message(self) -> str:
        return f"Checking {self.path}/*"


@dataclass
class ConfirmCleanPath(InfoLevel, NoFile):
    path: str
    code: str = "Z013"

    def message(self) -> str:
        return f"Cleaned {self.path}/*"


@dataclass
class ProtectedCleanPath(InfoLevel, NoFile):
    path: str
    code: str = "Z014"

    def message(self) -> str:
        return f"ERROR: not cleaning {self.path}/* because it is protected"


@dataclass
class FinishedCleanPaths(InfoLevel, NoFile):
    code: str = "Z015"

    def message(self) -> str:
        return "Finished cleaning all paths."


@dataclass
class OpenCommand(InfoLevel):
    open_cmd: str
    profiles_dir: str
    code: str = "Z016"

    def message(self) -> str:
        PROFILE_DIR_MESSAGE = """To view your profiles.yml file, run:

{open_cmd} {profiles_dir}"""
        message = PROFILE_DIR_MESSAGE.format(
            open_cmd=self.open_cmd, profiles_dir=self.profiles_dir
        )

        return message


@dataclass
class DepsNoPackagesFound(InfoLevel):
    code: str = "M013"

    def message(self) -> str:
        return "Warning: No packages were found in packages.yml"


@dataclass
class DepsStartPackageInstall(InfoLevel):
    package_name: str
    code: str = "M014"

    def message(self) -> str:
        return f"Installing {self.package_name}"


@dataclass
class DepsInstallInfo(InfoLevel):
    version_name: str
    code: str = "M015"

    def message(self) -> str:
        return f"  Installed from {self.version_name}"


@dataclass
class DepsUpdateAvailable(InfoLevel):
    version_latest: str
    code: str = "M016"

    def message(self) -> str:
        return f"  Updated version available: {self.version_latest}"


@dataclass
class DepsUTD(InfoLevel):
    code: str = "M017"

    def message(self) -> str:
        return "  Up to date!"


@dataclass
class DepsListSubdirectory(InfoLevel):
    subdirectory: str
    code: str = "M018"

    def message(self) -> str:
        return f"   and subdirectory {self.subdirectory}"


@dataclass
class DepsNotifyUpdatesAvailable(InfoLevel):
    packages: List[str]
    code: str = "M019"

    def message(self) -> str:
        return "Updates available for packages: {} \
                \nUpdate your versions in packages.yml, then run dbt deps".format(
            self.packages
        )


@dataclass
class DatabaseErrorRunning(InfoLevel):
    hook_type: str
    code: str = "E038"

    def message(self) -> str:
        return f"Database error while running {self.hook_type}"


@dataclass
class EmptyLine(InfoLevel):
    code: str = "Z017"

    def message(self) -> str:
        return ""


@dataclass
class HooksRunning(InfoLevel):
    num_hooks: int
    hook_type: str
    code: str = "E039"

    def message(self) -> str:
        plural = "hook" if self.num_hooks == 1 else "hooks"
        return f"Running {self.num_hooks} {self.hook_type} {plural}"


@dataclass
class HookFinished(InfoLevel):
    stat_line: str
    execution: str
    execution_time: float
    code: str = "E040"

    def message(self) -> str:
        return f"Finished running {self.stat_line}{self.execution} ({self.execution_time:0.2f}s)."


@dataclass
class WriteCatalogFailure(ErrorLevel):
    num_exceptions: int
    code: str = "E041"

    def message(self) -> str:
        return (
            f"dbt encountered {self.num_exceptions} failure{(self.num_exceptions != 1) * 's'} "
            "while writing the catalog"
        )


@dataclass
class CatalogWritten(InfoLevel):
    path: str
    code: str = "E042"

    def message(self) -> str:
        return f"Catalog written to {self.path}"


@dataclass
class CannotGenerateDocs(InfoLevel):
    code: str = "E043"

    def message(self) -> str:
        return "compile failed, cannot generate docs"


@dataclass
class BuildingCatalog(InfoLevel):
    code: str = "E044"

    def message(self) -> str:
        return "Building catalog"


@dataclass
class CompileComplete(InfoLevel):
    code: str = "Q002"

    def message(self) -> str:
        return "Done."


@dataclass
class FreshnessCheckComplete(InfoLevel):
    code: str = "Q003"

    def message(self) -> str:
        return "Done."


@dataclass
class ServingDocsPort(InfoLevel):
    address: str
    port: int
    code: str = "Z018"

    def message(self) -> str:
        return f"Serving docs at {self.address}:{self.port}"


@dataclass
class ServingDocsAccessInfo(InfoLevel):
    port: str
    code: str = "Z019"

    def message(self) -> str:
        return f"To access from your browser, navigate to:  http://localhost:{self.port}"


@dataclass
class ServingDocsExitInfo(InfoLevel):
    code: str = "Z020"

    def message(self) -> str:
        return "Press Ctrl+C to exit."


@dataclass
class SeedHeader(InfoLevel):
    header: str
    code: str = "Q004"

    def message(self) -> str:
        return self.header


@dataclass
class SeedHeaderSeparator(InfoLevel):
    len_header: int
    code: str = "Q005"

    def message(self) -> str:
        return "-" * self.len_header


@dataclass
class RunResultWarning(WarnLevel):
    resource_type: str
    node_name: str
    path: str
    code: str = "Z021"

    def message(self) -> str:
        info = "Warning"
        return ui.yellow(f"{info} in {self.resource_type} {self.node_name} ({self.path})")


@dataclass
class RunResultFailure(ErrorLevel):
    resource_type: str
    node_name: str
    path: str
    code: str = "Z022"

    def message(self) -> str:
        info = "Failure"
        return ui.red(f"{info} in {self.resource_type} {self.node_name} ({self.path})")


@dataclass
class StatsLine(InfoLevel):
    stats: Dict
    code: str = "Z023"

    def message(self) -> str:
        stats_line = "Done. PASS={pass} WARN={warn} ERROR={error} SKIP={skip} TOTAL={total}"
        return stats_line.format(**self.stats)


@dataclass
class RunResultError(ErrorLevel):
    msg: str
    code: str = "Z024"

    def message(self) -> str:
        return f"  {self.msg}"


@dataclass
class RunResultErrorNoMessage(ErrorLevel):
    status: str
    code: str = "Z025"

    def message(self) -> str:
        return f"  Status: {self.status}"


@dataclass
class SQLCompiledPath(InfoLevel):
    path: str
    code: str = "Z026"

    def message(self) -> str:
        return f"  compiled SQL at {self.path}"


@dataclass
class SQlRunnerException(ShowException, DebugLevel):
    exc: Exception
    code: str = "Q006"

    def message(self) -> str:
        return f"Got an exception: {self.exc}"


@dataclass
class CheckNodeTestFailure(InfoLevel):
    relation_name: str
    code: str = "Z027"

    def message(self) -> str:
        msg = f"select * from {self.relation_name}"
        border = "-" * len(msg)
        return f"  See test failures:\n  {border}\n  {msg}\n  {border}"


@dataclass
class FirstRunResultError(ErrorLevel):
    msg: str
    code: str = "Z028"

    def message(self) -> str:
        return ui.yellow(self.msg)


@dataclass
class AfterFirstRunResultError(ErrorLevel):
    msg: str
    code: str = "Z029"

    def message(self) -> str:
        return self.msg


@dataclass
class EndOfRunSummary(InfoLevel):
    num_errors: int
    num_warnings: int
    keyboard_interrupt: bool
    code: str = "Z030"

    def message(self) -> str:
        error_plural = pluralize(self.num_errors, "error")
        warn_plural = pluralize(self.num_warnings, "warning")
        if self.keyboard_interrupt:
            message = ui.yellow("Exited because of keyboard interrupt.")
        elif self.num_errors > 0:
            message = ui.red("Completed with {} and {}:".format(error_plural, warn_plural))
        elif self.num_warnings > 0:
            message = ui.yellow("Completed with {}:".format(warn_plural))
        else:
            message = ui.green("Completed successfully")
        return message


@dataclass
class PrintStartLine(InfoLevel, NodeInfo):
    description: str
    index: int
    total: int
    code: str = "Q033"

    def message(self) -> str:
        msg = f"START {self.description}"
        return format_fancy_output_line(msg=msg, status="RUN", index=self.index, total=self.total)


@dataclass
class PrintHookStartLine(InfoLevel, NodeInfo):
    statement: str
    index: int
    total: int
    code: str = "Q032"

    def message(self) -> str:
        msg = f"START hook: {self.statement}"
        return format_fancy_output_line(
            msg=msg, status="RUN", index=self.index, total=self.total, truncate=True
        )


@dataclass
class PrintHookEndLine(InfoLevel, NodeInfo):
    statement: str
    status: str
    index: int
    total: int
    execution_time: int
    code: str = "Q007"

    def message(self) -> str:
        msg = "OK hook: {}".format(self.statement)
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(self.status),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
            truncate=True,
        )


@dataclass
class SkippingDetails(InfoLevel, NodeInfo):
    resource_type: str
    schema: str
    node_name: str
    index: int
    total: int
    code: str = "Q034"

    def message(self) -> str:
        if self.resource_type in NodeType.refable():
            msg = f"SKIP relation {self.schema}.{self.node_name}"
        else:
            msg = f"SKIP {self.resource_type} {self.node_name}"
        return format_fancy_output_line(
            msg=msg, status=ui.yellow("SKIP"), index=self.index, total=self.total
        )


@dataclass
class PrintErrorTestResult(ErrorLevel, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    code: str = "Q008"

    def message(self) -> str:
        info = "ERROR"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(info),
            index=self.index,
            total=self.num_models,
            execution_time=self.execution_time,
        )


@dataclass
class PrintPassTestResult(InfoLevel, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    code: str = "Q009"

    def message(self) -> str:
        info = "PASS"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(info),
            index=self.index,
            total=self.num_models,
            execution_time=self.execution_time,
        )


@dataclass
class PrintWarnTestResult(WarnLevel, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    failures: int
    code: str = "Q010"

    def message(self) -> str:
        info = f"WARN {self.failures}"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.yellow(info),
            index=self.index,
            total=self.num_models,
            execution_time=self.execution_time,
        )


@dataclass
class PrintFailureTestResult(ErrorLevel, NodeInfo):
    name: str
    index: int
    num_models: int
    execution_time: int
    failures: int
    code: str = "Q011"

    def message(self) -> str:
        info = f"FAIL {self.failures}"
        msg = f"{info} {self.name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(info),
            index=self.index,
            total=self.num_models,
            execution_time=self.execution_time,
        )


@dataclass
class PrintSkipBecauseError(ErrorLevel):
    schema: str
    relation: str
    index: int
    total: int
    code: str = "Z034"

    def message(self) -> str:
        msg = f"SKIP relation {self.schema}.{self.relation} due to ephemeral model error"
        return format_fancy_output_line(
            msg=msg, status=ui.red("ERROR SKIP"), index=self.index, total=self.total
        )


@dataclass
class PrintModelErrorResultLine(ErrorLevel, NodeInfo):
    description: str
    status: str
    index: int
    total: int
    execution_time: int
    code: str = "Q035"

    def message(self) -> str:
        info = "ERROR creating"
        msg = f"{info} {self.description}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(self.status.upper()),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintModelResultLine(InfoLevel, NodeInfo):
    description: str
    status: str
    index: int
    total: int
    execution_time: int
    code: str = "Q012"

    def message(self) -> str:
        info = "OK created"
        msg = f"{info} {self.description}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(self.status),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintSnapshotErrorResultLine(ErrorLevel, NodeInfo):
    status: str
    description: str
    cfg: Dict[str, Any]
    index: int
    total: int
    execution_time: int
    code: str = "Q013"

    def message(self) -> str:
        info = "ERROR snapshotting"
        msg = "{info} {description}".format(info=info, description=self.description, **self.cfg)
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(self.status.upper()),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintSnapshotResultLine(InfoLevel, NodeInfo):
    status: str
    description: str
    cfg: Dict[str, Any]
    index: int
    total: int
    execution_time: int
    code: str = "Q014"

    def message(self) -> str:
        info = "OK snapshotted"
        msg = "{info} {description}".format(info=info, description=self.description, **self.cfg)
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(self.status),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintSeedErrorResultLine(ErrorLevel, NodeInfo):
    status: str
    index: int
    total: int
    execution_time: int
    schema: str
    relation: str
    code: str = "Q015"

    def message(self) -> str:
        info = "ERROR loading"
        msg = f"{info} seed file {self.schema}.{self.relation}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(self.status.upper()),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintSeedResultLine(InfoLevel, NodeInfo):
    status: str
    index: int
    total: int
    execution_time: int
    schema: str
    relation: str
    code: str = "Q016"

    def message(self) -> str:
        info = "OK loaded"
        msg = f"{info} seed file {self.schema}.{self.relation}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(self.status),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintHookEndErrorLine(ErrorLevel, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    code: str = "Q017"

    def message(self) -> str:
        info = "ERROR"
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(info),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintHookEndErrorStaleLine(ErrorLevel, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    code: str = "Q018"

    def message(self) -> str:
        info = "ERROR STALE"
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.red(info),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintHookEndWarnLine(WarnLevel, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    code: str = "Q019"

    def message(self) -> str:
        info = "WARN"
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.yellow(info),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintHookEndPassLine(InfoLevel, NodeInfo):
    source_name: str
    table_name: str
    index: int
    total: int
    execution_time: int
    code: str = "Q020"

    def message(self) -> str:
        info = "PASS"
        msg = f"{info} freshness of {self.source_name}.{self.table_name}"
        return format_fancy_output_line(
            msg=msg,
            status=ui.green(info),
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )


@dataclass
class PrintCancelLine(ErrorLevel):
    conn_name: str
    code: str = "Q021"

    def message(self) -> str:
        msg = "CANCEL query {}".format(self.conn_name)
        return format_fancy_output_line(msg=msg, status=ui.red("CANCEL"), index=None, total=None)


@dataclass
class DefaultSelector(InfoLevel):
    name: str
    code: str = "Q022"

    def message(self) -> str:
        return f"Using default selector {self.name}"


@dataclass
class NodeStart(DebugLevel, NodeInfo):
    unique_id: str
    code: str = "Q023"

    def message(self) -> str:
        return f"Began running node {self.unique_id}"


@dataclass
class NodeFinished(DebugLevel, NodeInfo):
    unique_id: str
    # The following isn't a RunResult class because we run into circular imports
    run_result: Dict[str, Any]
    code: str = "Q024"

    def message(self) -> str:
        return f"Finished running node {self.unique_id}"


@dataclass
class QueryCancelationUnsupported(InfoLevel):
    type: str
    code: str = "Q025"

    def message(self) -> str:
        msg = (
            f"The {self.type} adapter does not support query "
            "cancellation. Some queries may still be "
            "running!"
        )
        return ui.yellow(msg)


@dataclass
class ConcurrencyLine(InfoLevel):
    num_threads: int
    target_name: str
    code: str = "Q026"

    def message(self) -> str:
        return f"Concurrency: {self.num_threads} threads (target='{self.target_name}')"


@dataclass
class NodeCompiling(DebugLevel, NodeInfo):
    unique_id: str
    code: str = "Q030"

    def message(self) -> str:
        return f"Began compiling node {self.unique_id}"


@dataclass
class NodeExecuting(DebugLevel, NodeInfo):
    unique_id: str
    code: str = "Q031"

    def message(self) -> str:
        return f"Began executing node {self.unique_id}"


@dataclass
class StarterProjectPath(DebugLevel):
    dir: str
    code: str = "A017"

    def message(self) -> str:
        return f"Starter project path: {self.dir}"


@dataclass
class ConfigFolderDirectory(InfoLevel):
    dir: str
    code: str = "A018"

    def message(self) -> str:
        return f"Creating dbt configuration folder at {self.dir}"


@dataclass
class NoSampleProfileFound(InfoLevel):
    adapter: str
    code: str = "A019"

    def message(self) -> str:
        return f"No sample profile found for {self.adapter}."


@dataclass
class ProfileWrittenWithSample(InfoLevel):
    name: str
    path: str
    code: str = "A020"

    def message(self) -> str:
        return (
            f"Profile {self.name} written to {self.path} "
            "using target's sample configuration. Once updated, you'll be able to "
            "start developing with dbt."
        )


@dataclass
class ProfileWrittenWithTargetTemplateYAML(InfoLevel):
    name: str
    path: str
    code: str = "A021"

    def message(self) -> str:
        return (
            f"Profile {self.name} written to {self.path} using target's "
            "profile_template.yml and your supplied values. Run 'dbt debug' to "
            "validate the connection."
        )


@dataclass
class ProfileWrittenWithProjectTemplateYAML(InfoLevel):
    name: str
    path: str
    code: str = "A022"

    def message(self) -> str:
        return (
            f"Profile {self.name} written to {self.path} using project's "
            "profile_template.yml and your supplied values. Run 'dbt debug' to "
            "validate the connection."
        )


@dataclass
class SettingUpProfile(InfoLevel):
    code: str = "A023"

    def message(self) -> str:
        return "Setting up your profile."


@dataclass
class InvalidProfileTemplateYAML(InfoLevel):
    code: str = "A024"

    def message(self) -> str:
        return "Invalid profile_template.yml in project."


@dataclass
class ProjectNameAlreadyExists(InfoLevel):
    name: str
    code: str = "A025"

    def message(self) -> str:
        return f"A project called {self.name} already exists here."


@dataclass
class GetAddendum(InfoLevel):
    msg: str
    code: str = "A026"

    def message(self) -> str:
        return self.msg


@dataclass
class DepsSetDownloadDirectory(DebugLevel):
    path: str
    code: str = "A027"

    def message(self) -> str:
        return f"Set downloads directory='{self.path}'"


@dataclass
class EnsureGitInstalled(ErrorLevel):
    code: str = "Z036"

    def message(self) -> str:
        return (
            "Make sure git is installed on your machine. More "
            "information: "
            "https://docs.getdbt.com/docs/package-management"
        )


@dataclass
class DepsCreatingLocalSymlink(DebugLevel):
    code: str = "Z037"

    def message(self) -> str:
        return "  Creating symlink to local dependency."


@dataclass
class DepsSymlinkNotAvailable(DebugLevel):
    code: str = "Z038"

    def message(self) -> str:
        return "  Symlinks are not available on this OS, copying dependency."


@dataclass
class FoundStats(InfoLevel):
    stat_line: str
    code: str = "W006"

    def message(self) -> str:
        return f"Found {self.stat_line}"


# TODO: should this have NodeInfo on it?
@dataclass
class CompilingNode(DebugLevel):
    unique_id: str
    code: str = "Q027"

    def message(self) -> str:
        return f"Compiling {self.unique_id}"


@dataclass
class WritingInjectedSQLForNode(DebugLevel):
    unique_id: str
    code: str = "Q028"

    def message(self) -> str:
        return f'Writing injected SQL for node "{self.unique_id}"'


@dataclass
class DisableTracking(DebugLevel):
    code: str = "Z039"

    def message(self) -> str:
        return (
            "Error sending anonymous usage statistics. Disabling tracking for this execution. "
            "If you wish to permanently disable tracking, see: "
            "https://docs.getdbt.com/reference/global-configs#send-anonymous-usage-stats."
        )


@dataclass
class SendingEvent(DebugLevel):
    kwargs: str
    code: str = "Z040"

    def message(self) -> str:
        return f"Sending event: {self.kwargs}"


@dataclass
class SendEventFailure(DebugLevel):
    code: str = "Z041"

    def message(self) -> str:
        return "An error was encountered while trying to send an event"


@dataclass
class FlushEvents(DebugLevel):
    code: str = "Z042"

    def message(self) -> str:
        return "Flushing usage events"


@dataclass
class FlushEventsFailure(DebugLevel):
    code: str = "Z043"

    def message(self) -> str:
        return "An error was encountered while trying to flush usage events"


@dataclass
class TrackingInitializeFailure(ShowException, DebugLevel):
    code: str = "Z044"

    def message(self) -> str:
        return "Got an exception trying to initialize tracking"


@dataclass
class RetryExternalCall(DebugLevel):
    attempt: int
    max: int
    code: str = "M020"

    def message(self) -> str:
        return f"Retrying external call. Attempt: {self.attempt} Max attempts: {self.max}"


@dataclass
class GeneralWarningMsg(WarnLevel):
    msg: str
    log_fmt: str
    code: str = "Z046"

    def message(self) -> str:
        return self.log_fmt.format(self.msg) if self.log_fmt is not None else self.msg


@dataclass
class GeneralWarningException(WarnLevel):
    exc: Exception
    log_fmt: str
    code: str = "Z047"

    def message(self) -> str:
        return self.log_fmt.format(str(self.exc)) if self.log_fmt is not None else str(self.exc)


@dataclass
class EventBufferFull(WarnLevel):
    code: str = "Z048"

    def message(self) -> str:
        return (
            "Internal logging/event buffer full."
            "Earliest logs/events will be dropped as new ones are fired (FIFO)."
        )


@dataclass
class RecordRetryException(DebugLevel):
    exc: Exception
    code: str = "M021"

    def message(self) -> str:
        return f"External call exception: {self.exc}"


# since mypy doesn't run on every file we need to suggest to mypy that every
# class gets instantiated. But we don't actually want to run this code.
# making the conditional `if False` causes mypy to skip it as dead code so
# we need to skirt around that by computing something it doesn't check statically.
#
# TODO remove these lines once we run mypy everywhere.
if 1 == 0:
    MainReportVersion(v="")
    MainKeyboardInterrupt()
    MainEncounteredError(e=BaseException(""))
    MainStackTrace(stack_trace="")
    MainTrackingUserState(user_state="")
    ParsingStart()
    ParsingCompiling()
    ParsingWritingManifest()
    ParsingDone()
    ManifestDependenciesLoaded()
    ManifestLoaderCreated()
    ManifestLoaded()
    ManifestChecked()
    ManifestFlatGraphBuilt()
    ReportPerformancePath(path="")
    GitSparseCheckoutSubdirectory(subdir="")
    GitProgressCheckoutRevision(revision="")
    GitProgressUpdatingExistingDependency(dir="")
    GitProgressPullingNewDependency(dir="")
    GitNothingToDo(sha="")
    GitProgressUpdatedCheckoutRange(start_sha="", end_sha="")
    GitProgressCheckedOutAt(end_sha="")
    RegistryIndexProgressMakingGETRequest(url="")
    RegistryIndexProgressGETResponse(url="", resp_code=1234)
    RegistryProgressMakingGETRequest(url="")
    RegistryProgressGETResponse(url="", resp_code=1234)
    RegistryResponseUnexpectedType(response=""),
    RegistryResponseMissingTopKeys(response=""),
    RegistryResponseMissingNestedKeys(response=""),
    RegistryResponseExtraNestedKeys(response=""),
    SystemErrorRetrievingModTime(path="")
    SystemCouldNotWrite(path="", reason="", exc=Exception(""))
    SystemExecutingCmd(cmd=[""])
    SystemStdOutMsg(bmsg=b"")
    SystemStdErrMsg(bmsg=b"")
    SelectorReportInvalidSelector(valid_selectors="", spec_method="", raw_spec="")
    MacroEventInfo(msg="")
    MacroEventDebug(msg="")
    NewConnection(conn_type="", conn_name="")
    ConnectionReused(conn_name="")
    ConnectionLeftOpen(conn_name="")
    ConnectionClosed(conn_name="")
    RollbackFailed(conn_name="")
    ConnectionClosed2(conn_name="")
    ConnectionLeftOpen2(conn_name="")
    Rollback(conn_name="")
    CacheMiss(conn_name="", database="", schema="")
    ListRelations(database="", schema="", relations=[])
    ConnectionUsed(conn_type="", conn_name="")
    SQLQuery(conn_name="", sql="")
    SQLQueryStatus(status="", elapsed=0.1)
    SQLCommit(conn_name="")
    ColTypeChange(
        orig_type="", new_type="", table=_ReferenceKey(database="", schema="", identifier="")
    )
    SchemaCreation(relation=_ReferenceKey(database="", schema="", identifier=""))
    SchemaDrop(relation=_ReferenceKey(database="", schema="", identifier=""))
    UncachedRelation(
        dep_key=_ReferenceKey(database="", schema="", identifier=""),
        ref_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    AddLink(
        dep_key=_ReferenceKey(database="", schema="", identifier=""),
        ref_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    AddRelation(relation=_ReferenceKey(database="", schema="", identifier=""))
    DropMissingRelation(relation=_ReferenceKey(database="", schema="", identifier=""))
    DropCascade(
        dropped=_ReferenceKey(database="", schema="", identifier=""),
        consequences={_ReferenceKey(database="", schema="", identifier="")},
    )
    UpdateReference(
        old_key=_ReferenceKey(database="", schema="", identifier=""),
        new_key=_ReferenceKey(database="", schema="", identifier=""),
        cached_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    TemporaryRelation(key=_ReferenceKey(database="", schema="", identifier=""))
    RenameSchema(
        old_key=_ReferenceKey(database="", schema="", identifier=""),
        new_key=_ReferenceKey(database="", schema="", identifier=""),
    )
    DumpBeforeAddGraph(Lazy.defer(lambda: dict()))
    DumpAfterAddGraph(Lazy.defer(lambda: dict()))
    DumpBeforeRenameSchema(Lazy.defer(lambda: dict()))
    DumpAfterRenameSchema(Lazy.defer(lambda: dict()))
    AdapterImportError(exc=Exception())
    PluginLoadError()
    SystemReportReturnCode(returncode=0)
    NewConnectionOpening(connection_state="")
    TimingInfoCollected()
    MergedFromState(nbr_merged=0, sample=[])
    MissingProfileTarget(profile_name="", target_name="")
    InvalidVarsYAML()
    GenericTestFileParse(path="")
    MacroFileParse(path="")
    PartialParsingFullReparseBecauseOfError()
    PartialParsingFile(file_dict={})
    PartialParsingExceptionFile(file="")
    PartialParsingException(exc_info={})
    PartialParsingSkipParsing()
    PartialParsingMacroChangeStartFullParse()
    ManifestWrongMetadataVersion(version="")
    PartialParsingVersionMismatch(saved_version="", current_version="")
    PartialParsingFailedBecauseConfigChange()
    PartialParsingFailedBecauseProfileChange()
    PartialParsingFailedBecauseNewProjectDependency()
    PartialParsingFailedBecauseHashChanged()
    PartialParsingDeletedMetric(id="")
    ParsedFileLoadFailed(path="", exc=Exception(""))
    PartialParseSaveFileNotFound()
    StaticParserCausedJinjaRendering(path="")
    UsingExperimentalParser(path="")
    SampleFullJinjaRendering(path="")
    StaticParserFallbackJinjaRendering(path="")
    StaticParsingMacroOverrideDetected(path="")
    StaticParserSuccess(path="")
    StaticParserFailure(path="")
    ExperimentalParserSuccess(path="")
    ExperimentalParserFailure(path="")
    PartialParsingEnabled(deleted=0, added=0, changed=0)
    PartialParsingAddedFile(file_id="")
    PartialParsingDeletedFile(file_id="")
    PartialParsingUpdatedFile(file_id="")
    PartialParsingNodeMissingInSourceFile(source_file="")
    PartialParsingMissingNodes(file_id="")
    PartialParsingChildMapMissingUniqueID(unique_id="")
    PartialParsingUpdateSchemaFile(file_id="")
    PartialParsingDeletedSource(unique_id="")
    PartialParsingDeletedExposure(unique_id="")
    InvalidDisabledSourceInTestNode(msg="")
    InvalidRefInTestNode(msg="")
    RunningOperationCaughtError(exc=Exception(""))
    RunningOperationUncaughtError(exc=Exception(""))
    DbtProjectError()
    DbtProjectErrorException(exc=Exception(""))
    DbtProfileError()
    DbtProfileErrorException(exc=Exception(""))
    ProfileListTitle()
    ListSingleProfile(profile="")
    NoDefinedProfiles()
    ProfileHelpMessage()
    CatchableExceptionOnRun(exc=Exception(""))
    InternalExceptionOnRun(build_path="", exc=Exception(""))
    GenericExceptionOnRun(build_path="", unique_id="", exc=Exception(""))
    NodeConnectionReleaseError(node_name="", exc=Exception(""))
    CheckCleanPath(path="")
    ConfirmCleanPath(path="")
    ProtectedCleanPath(path="")
    FinishedCleanPaths()
    OpenCommand(open_cmd="", profiles_dir="")
    DepsNoPackagesFound()
    DepsStartPackageInstall(package_name="")
    DepsInstallInfo(version_name="")
    DepsUpdateAvailable(version_latest="")
    DepsListSubdirectory(subdirectory="")
    DepsNotifyUpdatesAvailable(packages=[])
    DatabaseErrorRunning(hook_type="")
    EmptyLine()
    HooksRunning(num_hooks=0, hook_type="")
    HookFinished(stat_line="", execution="", execution_time=0)
    WriteCatalogFailure(num_exceptions=0)
    CatalogWritten(path="")
    CannotGenerateDocs()
    BuildingCatalog()
    CompileComplete()
    FreshnessCheckComplete()
    ServingDocsPort(address="", port=0)
    ServingDocsAccessInfo(port="")
    ServingDocsExitInfo()
    SeedHeader(header="")
    SeedHeaderSeparator(len_header=0)
    RunResultWarning(resource_type="", node_name="", path="")
    RunResultFailure(resource_type="", node_name="", path="")
    StatsLine(stats={})
    RunResultError(msg="")
    RunResultErrorNoMessage(status="")
    SQLCompiledPath(path="")
    CheckNodeTestFailure(relation_name="")
    FirstRunResultError(msg="")
    AfterFirstRunResultError(msg="")
    EndOfRunSummary(num_errors=0, num_warnings=0, keyboard_interrupt=False)
    PrintStartLine(description="", index=0, total=0, node_info={})
    PrintHookStartLine(
        statement="",
        index=0,
        total=0,
        node_info={},
    )
    PrintHookEndLine(
        statement="",
        status="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    SkippingDetails(
        resource_type="",
        schema="",
        node_name="",
        index=0,
        total=0,
        node_info={},
    )
    PrintErrorTestResult(
        name="",
        index=0,
        num_models=0,
        execution_time=0,
        node_info={},
    )
    PrintPassTestResult(
        name="",
        index=0,
        num_models=0,
        execution_time=0,
        node_info={},
    )
    PrintWarnTestResult(
        name="",
        index=0,
        num_models=0,
        execution_time=0,
        failures=0,
        node_info={},
    )
    PrintFailureTestResult(
        name="",
        index=0,
        num_models=0,
        execution_time=0,
        failures=0,
        node_info={},
    )
    PrintSkipBecauseError(schema="", relation="", index=0, total=0)
    PrintModelErrorResultLine(
        description="",
        status="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintModelResultLine(
        description="",
        status="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintSnapshotErrorResultLine(
        status="",
        description="",
        cfg={},
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintSnapshotResultLine(
        status="",
        description="",
        cfg={},
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintSeedErrorResultLine(
        status="",
        index=0,
        total=0,
        execution_time=0,
        schema="",
        relation="",
        node_info={},
    )
    PrintSeedResultLine(
        status="",
        index=0,
        total=0,
        execution_time=0,
        schema="",
        relation="",
        node_info={},
    )
    PrintHookEndErrorLine(
        source_name="",
        table_name="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintHookEndErrorStaleLine(
        source_name="",
        table_name="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintHookEndWarnLine(
        source_name="",
        table_name="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintHookEndPassLine(
        source_name="",
        table_name="",
        index=0,
        total=0,
        execution_time=0,
        node_info={},
    )
    PrintCancelLine(conn_name="")
    DefaultSelector(name="")
    NodeStart(node_info={}, unique_id="")
    NodeFinished(node_info={}, unique_id="", run_result={})
    QueryCancelationUnsupported(type="")
    ConcurrencyLine(num_threads=0, target_name="")
    NodeCompiling(node_info={}, unique_id="")
    NodeExecuting(node_info={}, unique_id="")
    StarterProjectPath(dir="")
    ConfigFolderDirectory(dir="")
    NoSampleProfileFound(adapter="")
    ProfileWrittenWithSample(name="", path="")
    ProfileWrittenWithTargetTemplateYAML(name="", path="")
    ProfileWrittenWithProjectTemplateYAML(name="", path="")
    SettingUpProfile()
    InvalidProfileTemplateYAML()
    ProjectNameAlreadyExists(name="")
    GetAddendum(msg="")
    DepsSetDownloadDirectory(path="")
    EnsureGitInstalled()
    DepsCreatingLocalSymlink()
    DepsSymlinkNotAvailable()
    FoundStats(stat_line="")
    CompilingNode(unique_id="")
    WritingInjectedSQLForNode(unique_id="")
    DisableTracking()
    SendingEvent(kwargs="")
    SendEventFailure()
    FlushEvents()
    FlushEventsFailure()
    TrackingInitializeFailure()
    RetryExternalCall(attempt=0, max=0)
    GeneralWarningMsg(msg="", log_fmt="")
    GeneralWarningException(exc=Exception(""), log_fmt="")
    EventBufferFull()
    RecordRetryException(exc=Exception(""))



from typing import Any, Dict, Union

from dbt.exceptions import (
    doc_invalid_args,
    doc_target_not_found,
)
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.graph.compiled import CompileResultNode
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.graph.parsed import ParsedMacro

from dbt.context.base import contextmember
from dbt.context.configured import SchemaYamlContext


class DocsRuntimeContext(SchemaYamlContext):
    def __init__(
        self,
        config: RuntimeConfig,
        node: Union[ParsedMacro, CompileResultNode],
        manifest: Manifest,
        current_project: str,
    ) -> None:
        super().__init__(config, current_project, None)
        self.node = node
        self.manifest = manifest

    @contextmember
    def doc(self, *args: str) -> str:
        """The `doc` function is used to reference docs blocks in schema.yml
        files. It is analogous to the `ref` function. For more information,
        consult the Documentation guide.

        > orders.md:

            {% docs orders %}
            # docs
            - go
            - here
            {% enddocs %}

        > schema.yml

            version: 2
                models:
                  - name: orders
                    description: "{{ doc('orders') }}"
        """
        # when you call doc(), this is what happens at runtime
        if len(args) == 1:
            doc_package_name = None
            doc_name = args[0]
        elif len(args) == 2:
            doc_package_name, doc_name = args
        else:
            doc_invalid_args(self.node, args)

        # ParsedDocumentation
        target_doc = self.manifest.resolve_doc(
            doc_name,
            doc_package_name,
            self._project_name,
            self.node.package_name,
        )
        if target_doc:
            file_id = target_doc.file_id
            if file_id in self.manifest.files:
                source_file = self.manifest.files[file_id]
                # TODO CT-211
                source_file.add_node(self.node.unique_id)  # type: ignore[union-attr]
        else:
            doc_target_not_found(self.node, doc_name, doc_package_name)

        return target_doc.block_contents


def generate_runtime_docs_context(
    config: RuntimeConfig,
    target: Any,
    manifest: Manifest,
    current_project: str,
) -> Dict[str, Any]:
    ctx = DocsRuntimeContext(config, target, manifest, current_project)
    # This is not a Mashumaro to_dict call
    return ctx.to_dict()



import itertools
from pathlib import Path
from typing import Iterable, Dict, Optional, Set, Any
from dbt.adapters.factory import get_adapter
from dbt.config import RuntimeConfig
from dbt.context.context_config import (
    BaseContextConfigGenerator,
    ContextConfigGenerator,
    UnrenderedConfigGenerator,
)
from dbt.contracts.graph.manifest import Manifest, SourceKey
from dbt.contracts.graph.model_config import SourceConfig
from dbt.contracts.graph.parsed import (
    UnpatchedSourceDefinition,
    ParsedSourceDefinition,
    ParsedGenericTestNode,
)
from dbt.contracts.graph.unparsed import (
    UnparsedSourceDefinition,
    SourcePatch,
    SourceTablePatch,
    UnparsedSourceTableDefinition,
    FreshnessThreshold,
    UnparsedColumn,
    Time,
)
from dbt.exceptions import warn_or_error, InternalException
from dbt.node_types import NodeType

from dbt.parser.schemas import SchemaParser, ParserRef
from dbt import ui


# An UnparsedSourceDefinition is taken directly from the yaml
# file. It can affect multiple tables, all of which will eventually
# have their own source node. An UnparsedSourceDefinition will
# generate multiple UnpatchedSourceDefinition nodes (one per
# table) in the SourceParser.add_source_definitions. The
# SourcePatcher takes an UnparsedSourceDefinition and the
# SourcePatch and produces a ParsedSourceDefinition. Each
# SourcePatch can be applied to multiple UnpatchedSourceDefinitions.
class SourcePatcher:
    def __init__(
        self,
        root_project: RuntimeConfig,
        manifest: Manifest,
    ) -> None:
        self.root_project = root_project
        self.manifest = manifest
        self.schema_parsers: Dict[str, SchemaParser] = {}
        self.patches_used: Dict[SourceKey, Set[str]] = {}
        self.sources: Dict[str, ParsedSourceDefinition] = {}

    # This method calls the 'parse_source' method which takes
    # the UnpatchedSourceDefinitions in the manifest and combines them
    # with SourcePatches to produce ParsedSourceDefinitions.
    def construct_sources(self) -> None:
        for unique_id, unpatched in self.manifest.sources.items():
            schema_file = self.manifest.files[unpatched.file_id]
            if isinstance(unpatched, ParsedSourceDefinition):
                # In partial parsing, there will be ParsedSourceDefinitions
                # which must be retained.
                self.sources[unpatched.unique_id] = unpatched
                continue
            # returns None if there is no patch
            patch = self.get_patch_for(unpatched)  # type: ignore[unreachable] # CT-564 / GH 5169

            # returns unpatched if there is no patch
            patched = self.patch_source(unpatched, patch)

            # now use the patched UnpatchedSourceDefinition to extract test data.
            for test in self.get_source_tests(patched):
                if test.config.enabled:
                    self.manifest.add_node_nofile(test)
                else:
                    self.manifest.add_disabled_nofile(test)
                # save the test unique_id in the schema_file, so we can
                # process in partial parsing
                test_from = {"key": "sources", "name": patched.source.name}
                schema_file.add_test(test.unique_id, test_from)

            # Convert UnpatchedSourceDefinition to a ParsedSourceDefinition
            parsed = self.parse_source(patched)
            if parsed.config.enabled:
                self.sources[unique_id] = parsed
            else:
                self.manifest.add_disabled_nofile(parsed)

        self.warn_unused()

    def patch_source(
        self,
        unpatched: UnpatchedSourceDefinition,
        patch: Optional[SourcePatch],
    ) -> UnpatchedSourceDefinition:

        # This skips patching if no patch exists because of the
        # performance overhead of converting to and from dicts
        if patch is None:
            return unpatched

        source_dct = unpatched.source.to_dict(omit_none=True)
        table_dct = unpatched.table.to_dict(omit_none=True)
        patch_path: Optional[Path] = None

        source_table_patch: Optional[SourceTablePatch] = None

        if patch is not None:
            source_table_patch = patch.get_table_named(unpatched.table.name)
            source_dct.update(patch.to_patch_dict())
            patch_path = patch.path

        if source_table_patch is not None:
            table_dct.update(source_table_patch.to_patch_dict())

        source = UnparsedSourceDefinition.from_dict(source_dct)
        table = UnparsedSourceTableDefinition.from_dict(table_dct)
        return unpatched.replace(source=source, table=table, patch_path=patch_path)

    # This converts an UnpatchedSourceDefinition to a ParsedSourceDefinition
    def parse_source(self, target: UnpatchedSourceDefinition) -> ParsedSourceDefinition:
        source = target.source
        table = target.table
        refs = ParserRef.from_target(table)
        unique_id = target.unique_id
        description = table.description or ""
        meta = table.meta or {}
        source_description = source.description or ""
        loaded_at_field = table.loaded_at_field or source.loaded_at_field

        freshness = merge_freshness(source.freshness, table.freshness)
        quoting = source.quoting.merged(table.quoting)
        # path = block.path.original_file_path
        source_meta = source.meta or {}

        # make sure we don't do duplicate tags from source + table
        tags = sorted(set(itertools.chain(source.tags, table.tags)))

        config = self._generate_source_config(
            target=target,
            rendered=True,
        )

        unrendered_config = self._generate_source_config(
            target=target,
            rendered=False,
        )

        if not isinstance(config, SourceConfig):
            raise InternalException(
                f"Calculated a {type(config)} for a source, but expected " f"a SourceConfig"
            )

        default_database = self.root_project.credentials.database

        parsed_source = ParsedSourceDefinition(
            package_name=target.package_name,
            database=(source.database or default_database),
            schema=(source.schema or source.name),
            identifier=(table.identifier or table.name),
            root_path=target.root_path,
            path=target.path,
            original_file_path=target.original_file_path,
            columns=refs.column_info,
            unique_id=unique_id,
            name=table.name,
            description=description,
            external=table.external,
            source_name=source.name,
            source_description=source_description,
            source_meta=source_meta,
            meta=meta,
            loader=source.loader,
            loaded_at_field=loaded_at_field,
            freshness=freshness,
            quoting=quoting,
            resource_type=NodeType.Source,
            fqn=target.fqn,
            tags=tags,
            config=config,
            unrendered_config=unrendered_config,
        )

        # relation name is added after instantiation because the adapter does
        # not provide the relation name for a UnpatchedSourceDefinition object
        parsed_source.relation_name = self._get_relation_name(parsed_source)
        return parsed_source

    # This code uses the SchemaParser because it shares the '_parse_generic_test'
    # code. It might be nice to separate out the generic test code
    # and make it common to the schema parser and source patcher.
    def get_schema_parser_for(self, package_name: str) -> "SchemaParser":
        if package_name in self.schema_parsers:
            schema_parser = self.schema_parsers[package_name]
        else:
            all_projects = self.root_project.load_dependencies()
            project = all_projects[package_name]
            schema_parser = SchemaParser(project, self.manifest, self.root_project)
            self.schema_parsers[package_name] = schema_parser
        return schema_parser

    def get_source_tests(
        self, target: UnpatchedSourceDefinition
    ) -> Iterable[ParsedGenericTestNode]:
        for test, column in target.get_tests():
            yield self.parse_source_test(
                target=target,
                test=test,
                column=column,
            )

    def get_patch_for(
        self,
        unpatched: UnpatchedSourceDefinition,
    ) -> Optional[SourcePatch]:
        if isinstance(unpatched, ParsedSourceDefinition):  # type: ignore[unreachable] # CT-564 / GH 5169
            return None  # type: ignore[unreachable] # CT-564 / GH 5169
        key = (unpatched.package_name, unpatched.source.name)
        patch: Optional[SourcePatch] = self.manifest.source_patches.get(key)
        if patch is None:
            return None
        if key not in self.patches_used:
            # mark the key as used
            self.patches_used[key] = set()
        if patch.get_table_named(unpatched.table.name) is not None:
            self.patches_used[key].add(unpatched.table.name)
        return patch

    # This calls _parse_generic_test in the SchemaParser
    def parse_source_test(
        self,
        target: UnpatchedSourceDefinition,
        test: Dict[str, Any],
        column: Optional[UnparsedColumn],
    ) -> ParsedGenericTestNode:
        column_name: Optional[str]
        if column is None:
            column_name = None
        else:
            column_name = column.name
            should_quote = column.quote or (column.quote is None and target.quote_columns)
            if should_quote:
                column_name = get_adapter(self.root_project).quote(column_name)

        tags_sources = [target.source.tags, target.table.tags]
        if column is not None:
            tags_sources.append(column.tags)
        tags = list(itertools.chain.from_iterable(tags_sources))

        # TODO: make the generic_test code common so we don't need to
        # create schema parsers to handle the tests
        schema_parser = self.get_schema_parser_for(target.package_name)
        node = schema_parser._parse_generic_test(
            target=target,
            test=test,
            tags=tags,
            column_name=column_name,
            schema_file_id=target.file_id,
        )
        return node

    def _generate_source_config(self, target: UnpatchedSourceDefinition, rendered: bool):
        generator: BaseContextConfigGenerator
        if rendered:
            generator = ContextConfigGenerator(self.root_project)
        else:
            generator = UnrenderedConfigGenerator(self.root_project)

        # configs with precendence set
        precedence_configs = dict()
        # first apply source configs
        precedence_configs.update(target.source.config)
        # then overrite anything that is defined on source tables
        # this is not quite complex enough for configs that can be set as top-level node keys, but
        # it works while source configs can only include `enabled`.
        precedence_configs.update(target.table.config)

        return generator.calculate_node_config(
            config_call_dict={},
            fqn=target.fqn,
            resource_type=NodeType.Source,
            project_name=target.package_name,
            base=False,
            patch_config_dict=precedence_configs,
        )

    def _get_relation_name(self, node: ParsedSourceDefinition):
        adapter = get_adapter(self.root_project)
        relation_cls = adapter.Relation
        return str(relation_cls.create_from(self.root_project, node))

    def warn_unused(self) -> None:
        unused_tables: Dict[SourceKey, Optional[Set[str]]] = {}
        for patch in self.manifest.source_patches.values():
            key = (patch.overrides, patch.name)
            if key not in self.patches_used:
                unused_tables[key] = None
            elif patch.tables is not None:
                table_patches = {t.name for t in patch.tables}
                unused = table_patches - self.patches_used[key]
                # don't add unused tables, the
                if unused:
                    # because patches are required to be unique, we can safely
                    # write without looking
                    unused_tables[key] = unused

        if unused_tables:
            msg = self.get_unused_msg(unused_tables)
            warn_or_error(msg, log_fmt=ui.warning_tag("{}"))

        self.manifest.source_patches = {}

    def get_unused_msg(
        self,
        unused_tables: Dict[SourceKey, Optional[Set[str]]],
    ) -> str:
        msg = [
            "During parsing, dbt encountered source overrides that had no " "target:",
        ]
        for key, table_names in unused_tables.items():
            patch = self.manifest.source_patches[key]
            patch_name = f"{patch.overrides}.{patch.name}"
            if table_names is None:
                msg.append(f"  - Source {patch_name} (in {patch.path})")
            else:
                for table_name in sorted(table_names):
                    msg.append(f"  - Source table {patch_name}.{table_name} " f"(in {patch.path})")
        msg.append("")
        return "\n".join(msg)


def merge_freshness_time_thresholds(
    base: Optional[Time], update: Optional[Time]
) -> Optional[Time]:
    if base and update:
        return base.merged(update)
    elif update is None:
        return None
    else:
        return update or base


def merge_freshness(
    base: Optional[FreshnessThreshold], update: Optional[FreshnessThreshold]
) -> Optional[FreshnessThreshold]:
    if base is not None and update is not None:
        merged_freshness = base.merged(update)
        # merge one level deeper the error_after and warn_after thresholds
        merged_error_after = merge_freshness_time_thresholds(base.error_after, update.error_after)
        merged_warn_after = merge_freshness_time_thresholds(base.warn_after, update.warn_after)

        merged_freshness.error_after = merged_error_after
        merged_freshness.warn_after = merged_warn_after
        return merged_freshness
    elif base is None and update is not None:
        return update
    else:
        return None



# never name this package "types", or mypy will crash in ugly ways

# necessary for annotating constructors
from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Tuple, AbstractSet, Union

from dbt.dataclass_schema import (
    dbtClassMixin,
    ValidationError,
    StrEnum,
)
from hologram import FieldEncoder, JsonDict
from mashumaro.types import SerializableType
from typing import Callable, cast, Generic, Optional, TypeVar


class Port(int, SerializableType):
    @classmethod
    def _deserialize(cls, value: Union[int, str]) -> "Port":
        try:
            value = int(value)
        except ValueError:
            raise ValidationError(f"Cannot encode {value} into port number")

        return Port(value)

    def _serialize(self) -> int:
        return self


class PortEncoder(FieldEncoder):
    @property
    def json_schema(self):
        return {"type": "integer", "minimum": 0, "maximum": 65535}


class TimeDeltaFieldEncoder(FieldEncoder[timedelta]):
    """Encodes timedeltas to dictionaries"""

    def to_wire(self, value: timedelta) -> float:
        return value.total_seconds()

    def to_python(self, value) -> timedelta:
        if isinstance(value, timedelta):
            return value
        try:
            return timedelta(seconds=value)
        except TypeError:
            raise ValidationError("cannot encode {} into timedelta".format(value)) from None

    @property
    def json_schema(self) -> JsonDict:
        return {"type": "number"}


class PathEncoder(FieldEncoder):
    def to_wire(self, value: Path) -> str:
        return str(value)

    def to_python(self, value) -> Path:
        if isinstance(value, Path):
            return value
        try:
            return Path(value)
        except TypeError:
            raise ValidationError("cannot encode {} into timedelta".format(value)) from None

    @property
    def json_schema(self) -> JsonDict:
        return {"type": "string"}


class NVEnum(StrEnum):
    novalue = "novalue"

    def __eq__(self, other):
        return isinstance(other, NVEnum)


@dataclass
class NoValue(dbtClassMixin):
    """Sometimes, you want a way to say none that isn't None"""

    novalue: NVEnum = NVEnum.novalue


dbtClassMixin.register_field_encoders(
    {
        Port: PortEncoder(),
        timedelta: TimeDeltaFieldEncoder(),
        Path: PathEncoder(),
    }
)


FQNPath = Tuple[str, ...]
PathSet = AbstractSet[FQNPath]

T = TypeVar("T")


# A data type for representing lazily evaluated values.
#
# usage:
# x = Lazy.defer(lambda: expensive_fn())
# y = x.force()
#
# inspired by the purescript data type
# https://pursuit.purescript.org/packages/purescript-lazy/5.0.0/docs/Data.Lazy
@dataclass
class Lazy(Generic[T]):
    _f: Callable[[], T]
    memo: Optional[T] = None

    # constructor for lazy values
    @classmethod
    def defer(cls, f: Callable[[], T]) -> Lazy[T]:
        return Lazy(f)

    # workaround for open mypy issue:
    # https://github.com/python/mypy/issues/6910
    def _typed_eval_f(self) -> T:
        return cast(Callable[[], T], getattr(self, "_f"))()

    # evaluates the function if the value has not been memoized already
    def force(self) -> T:
        if self.memo is None:
            self.memo = self._typed_eval_f()
        return self.memo


# This class is used in to_target_dict, so that accesses to missing keys
# will return an empty string instead of Undefined
class DictDefaultEmptyStr(dict):
    def __getitem__(self, key):
        return dict.get(self, key, "")



from dbt.node_types import NodeType


class MetricReference(object):
    def __init__(self, metric_name, package_name=None):
        self.metric_name = metric_name
        self.package_name = package_name

    def __str__(self):
        return f"{self.metric_name}"


class ResolvedMetricReference(MetricReference):
    """
    Simple proxy over a ParsedMetric which delegates property
    lookups to the underlying node. Also adds helper functions
    for working with metrics (ie. __str__ and templating functions)
    """

    def __init__(self, node, manifest, Relation):
        super().__init__(node.name, node.package_name)
        self.node = node
        self.manifest = manifest
        self.Relation = Relation

    def __getattr__(self, key):
        return getattr(self.node, key)

    def __str__(self):
        return f"{self.node.name}"

    @classmethod
    def parent_metrics(cls, metric_node, manifest):
        yield metric_node

        for parent_unique_id in metric_node.depends_on.nodes:
            node = manifest.metrics.get(parent_unique_id)
            if node and node.resource_type == NodeType.Metric:
                yield from cls.parent_metrics(node, manifest)

    def parent_models(self):
        in_scope_metrics = list(self.parent_metrics(self.node, self.manifest))

        to_return = {
            "base": [],
            "derived": [],
        }
        for metric in in_scope_metrics:
            if metric.type == "expression":
                to_return["derived"].append(
                    {"metric_source": None, "metric": metric, "is_derived": True}
                )
            else:
                for node_unique_id in metric.depends_on.nodes:
                    node = self.manifest.nodes.get(node_unique_id)
                    if node and node.resource_type in NodeType.refable():
                        to_return["base"].append(
                            {
                                "metric_relation_node": node,
                                "metric_relation": self.Relation.create(
                                    database=node.database,
                                    schema=node.schema,
                                    identifier=node.alias,
                                ),
                                "metric": metric,
                                "is_derived": False,
                            }
                        )

        return to_return



from distutils.util import strtobool

from dataclasses import dataclass
from dbt.utils import _coerce_decimal
from dbt.events.format import pluralize
from dbt.dataclass_schema import dbtClassMixin
import threading
from typing import Union

from .compile import CompileRunner
from .run import RunTask

from dbt.contracts.graph.compiled import (
    CompiledSingularTestNode,
    CompiledGenericTestNode,
    CompiledTestNode,
)
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.results import TestStatus, PrimitiveDict, RunResult
from dbt.context.providers import generate_runtime_model_context
from dbt.clients.jinja import MacroGenerator
from dbt.events.functions import fire_event
from dbt.events.types import (
    PrintErrorTestResult,
    PrintPassTestResult,
    PrintWarnTestResult,
    PrintFailureTestResult,
    PrintStartLine,
)
from dbt.exceptions import InternalException, invalid_bool_error, missing_materialization
from dbt.graph import (
    ResourceTypeSelector,
)
from dbt.node_types import NodeType
from dbt import flags


@dataclass
class TestResultData(dbtClassMixin):
    failures: int
    should_warn: bool
    should_error: bool

    @classmethod
    def validate(cls, data):
        data["should_warn"] = cls.convert_bool_type(data["should_warn"])
        data["should_error"] = cls.convert_bool_type(data["should_error"])
        super().validate(data)

    def convert_bool_type(field) -> bool:
        # if it's type string let python decide if it's a valid value to convert to bool
        if isinstance(field, str):
            try:
                return bool(strtobool(field))  # type: ignore
            except ValueError:
                raise invalid_bool_error(field, "get_test_sql")

        # need this so we catch both true bools and 0/1
        return bool(field)


class TestRunner(CompileRunner):
    def describe_node(self):
        node_name = self.node.name
        return "test {}".format(node_name)

    def print_result_line(self, result):
        model = result.node

        if result.status == TestStatus.Error:
            fire_event(
                PrintErrorTestResult(
                    name=model.name,
                    index=self.node_index,
                    num_models=self.num_nodes,
                    execution_time=result.execution_time,
                    node_info=model.node_info,
                )
            )
        elif result.status == TestStatus.Pass:
            fire_event(
                PrintPassTestResult(
                    name=model.name,
                    index=self.node_index,
                    num_models=self.num_nodes,
                    execution_time=result.execution_time,
                    node_info=model.node_info,
                )
            )
        elif result.status == TestStatus.Warn:
            fire_event(
                PrintWarnTestResult(
                    name=model.name,
                    index=self.node_index,
                    num_models=self.num_nodes,
                    execution_time=result.execution_time,
                    failures=result.failures,
                    node_info=model.node_info,
                )
            )
        elif result.status == TestStatus.Fail:
            fire_event(
                PrintFailureTestResult(
                    name=model.name,
                    index=self.node_index,
                    num_models=self.num_nodes,
                    execution_time=result.execution_time,
                    failures=result.failures,
                    node_info=model.node_info,
                )
            )
        else:
            raise RuntimeError("unexpected status: {}".format(result.status))

    def print_start_line(self):
        fire_event(
            PrintStartLine(
                description=self.describe_node(),
                index=self.node_index,
                total=self.num_nodes,
                node_info=self.node.node_info,
            )
        )

    def before_execute(self):
        self.print_start_line()

    def execute_test(
        self, test: Union[CompiledSingularTestNode, CompiledGenericTestNode], manifest: Manifest
    ) -> TestResultData:
        context = generate_runtime_model_context(test, self.config, manifest)

        materialization_macro = manifest.find_materialization_macro_by_name(
            self.config.project_name, test.get_materialization(), self.adapter.type()
        )

        if materialization_macro is None:
            missing_materialization(test, self.adapter.type())

        if "config" not in context:
            raise InternalException(
                "Invalid materialization context generated, missing config: {}".format(context)
            )

        # generate materialization macro
        macro_func = MacroGenerator(materialization_macro, context)
        # execute materialization macro
        macro_func()
        # load results from context
        # could eventually be returned directly by materialization
        result = context["load_result"]("main")
        table = result["table"]
        num_rows = len(table.rows)
        if num_rows != 1:
            raise InternalException(
                f"dbt internally failed to execute {test.unique_id}: "
                f"Returned {num_rows} rows, but expected "
                f"1 row"
            )
        num_cols = len(table.columns)
        if num_cols != 3:
            raise InternalException(
                f"dbt internally failed to execute {test.unique_id}: "
                f"Returned {num_cols} columns, but expected "
                f"3 columns"
            )

        test_result_dct: PrimitiveDict = dict(
            zip(
                [column_name.lower() for column_name in table.column_names],
                map(_coerce_decimal, table.rows[0]),
            )
        )
        TestResultData.validate(test_result_dct)
        return TestResultData.from_dict(test_result_dct)

    def execute(self, test: CompiledTestNode, manifest: Manifest):
        result = self.execute_test(test, manifest)

        severity = test.config.severity.upper()
        thread_id = threading.current_thread().name
        num_errors = pluralize(result.failures, "result")
        status = None
        message = None
        failures = 0
        if severity == "ERROR" and result.should_error:
            status = TestStatus.Fail
            message = f"Got {num_errors}, configured to fail if {test.config.error_if}"
            failures = result.failures
        elif result.should_warn:
            if flags.WARN_ERROR:
                status = TestStatus.Fail
                message = f"Got {num_errors}, configured to fail if {test.config.warn_if}"
            else:
                status = TestStatus.Warn
                message = f"Got {num_errors}, configured to warn if {test.config.warn_if}"
            failures = result.failures
        else:
            status = TestStatus.Pass

        return RunResult(
            node=test,
            status=status,
            timing=[],
            thread_id=thread_id,
            execution_time=0,
            message=message,
            adapter_response={},
            failures=failures,
        )

    def after_execute(self, result):
        self.print_result_line(result)


class TestSelector(ResourceTypeSelector):
    def __init__(self, graph, manifest, previous_state):
        super().__init__(
            graph=graph,
            manifest=manifest,
            previous_state=previous_state,
            resource_types=[NodeType.Test],
        )


class TestTask(RunTask):
    """
    Testing:
        Read schema files + custom data tests and validate that
        constraints are satisfied.
    """

    __test__ = False

    def raise_on_first_error(self):
        return False

    def get_node_selector(self) -> TestSelector:
        if self.manifest is None or self.graph is None:
            raise InternalException("manifest and graph must be set to get perform node selection")
        return TestSelector(
            graph=self.graph,
            manifest=self.manifest,
            previous_state=self.previous_state,
        )

    def get_runner_type(self, _):
        return TestRunner



from typing import List
from dbt.logger import log_cache_events, log_manager

import argparse
import os.path
import sys
import traceback
import warnings
from contextlib import contextmanager
from pathlib import Path

import dbt.version
from dbt.events.functions import fire_event, setup_event_logger
from dbt.events.types import (
    MainEncounteredError,
    MainKeyboardInterrupt,
    MainReportVersion,
    MainReportArgs,
    MainTrackingUserState,
    MainStackTrace,
)
import dbt.flags as flags
import dbt.task.build as build_task
import dbt.task.clean as clean_task
import dbt.task.compile as compile_task
import dbt.task.debug as debug_task
import dbt.task.deps as deps_task
import dbt.task.freshness as freshness_task
import dbt.task.generate as generate_task
import dbt.task.init as init_task
import dbt.task.list as list_task
import dbt.task.parse as parse_task
import dbt.task.run as run_task
import dbt.task.run_operation as run_operation_task
import dbt.task.seed as seed_task
import dbt.task.serve as serve_task
import dbt.task.snapshot as snapshot_task
import dbt.task.test as test_task
from dbt.profiler import profiler
from dbt.adapters.factory import reset_adapters, cleanup_connections

import dbt.tracking

from dbt.utils import ExitCodes, args_to_dict
from dbt.config.profile import DEFAULT_PROFILES_DIR, read_user_config
from dbt.exceptions import InternalException, NotImplementedException, FailedToConnectException


class DBTVersion(argparse.Action):
    """This is very similar to the built-in argparse._Version action,
    except it just calls dbt.version.get_version_information().
    """

    def __init__(
        self,
        option_strings,
        version=None,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help="show program's version number and exit",
    ):
        super().__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        formatter = argparse.RawTextHelpFormatter(prog=parser.prog)
        formatter.add_text(dbt.version.get_version_information())
        parser.exit(message=formatter.format_help())


class DBTArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register("action", "dbtversion", DBTVersion)

    def add_optional_argument_inverse(
        self,
        name,
        *,
        enable_help=None,
        disable_help=None,
        dest=None,
        no_name=None,
        default=None,
    ):
        mutex_group = self.add_mutually_exclusive_group()
        if not name.startswith("--"):
            raise InternalException(
                'cannot handle optional argument without "--" prefix: ' f'got "{name}"'
            )
        if dest is None:
            dest_name = name[2:].replace("-", "_")
        else:
            dest_name = dest

        if no_name is None:
            no_name = f"--no-{name[2:]}"

        mutex_group.add_argument(
            name,
            action="store_const",
            const=True,
            dest=dest_name,
            default=default,
            help=enable_help,
        )

        mutex_group.add_argument(
            f"--no-{name[2:]}",
            action="store_const",
            const=False,
            dest=dest_name,
            default=default,
            help=disable_help,
        )

        return mutex_group


def main(args=None):
    # Logbook warnings are ignored so we don't have to fork logbook to support python 3.10.
    # This _only_ works for regular cli invocations.
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="logbook")
    if args is None:
        args = sys.argv[1:]
    with log_manager.applicationbound():
        try:
            results, succeeded = handle_and_check(args)
            if succeeded:
                exit_code = ExitCodes.Success.value
            else:
                exit_code = ExitCodes.ModelError.value

        except KeyboardInterrupt:
            # if the logger isn't configured yet, it will use the default logger
            fire_event(MainKeyboardInterrupt())
            exit_code = ExitCodes.UnhandledError.value

        # This can be thrown by eg. argparse
        except SystemExit as e:
            exit_code = e.code

        except BaseException as e:
            fire_event(MainEncounteredError(e=str(e)))
            fire_event(MainStackTrace(stack_trace=traceback.format_exc()))
            exit_code = ExitCodes.UnhandledError.value

    sys.exit(exit_code)


# here for backwards compatibility
def handle(args):
    res, success = handle_and_check(args)
    return res


@contextmanager
def adapter_management():
    reset_adapters()
    try:
        yield
    finally:
        cleanup_connections()


def handle_and_check(args):
    with log_manager.applicationbound():
        parsed = parse_args(args)

        # Set flags from args, user config, and env vars
        user_config = read_user_config(flags.PROFILES_DIR)  # This is read again later
        flags.set_from_args(parsed, user_config)
        dbt.tracking.initialize_from_flags()
        # Set log_format from flags
        parsed.cls.set_log_format()

        # we've parsed the args and set the flags - we can now decide if we're debug or not
        if flags.DEBUG:
            log_manager.set_debug()

        profiler_enabled = False

        if parsed.record_timing_info:
            profiler_enabled = True

        with profiler(enable=profiler_enabled, outfile=parsed.record_timing_info):

            with adapter_management():

                task, res = run_from_args(parsed)
                success = task.interpret_results(res)

            return res, success


@contextmanager
def track_run(task):
    dbt.tracking.track_invocation_start(config=task.config, args=task.args)
    try:
        yield
        dbt.tracking.track_invocation_end(config=task.config, args=task.args, result_type="ok")
    except (NotImplementedException, FailedToConnectException) as e:
        fire_event(MainEncounteredError(e=str(e)))
        dbt.tracking.track_invocation_end(config=task.config, args=task.args, result_type="error")
    except Exception:
        dbt.tracking.track_invocation_end(config=task.config, args=task.args, result_type="error")
        raise
    finally:
        dbt.tracking.flush()


def run_from_args(parsed):
    log_cache_events(getattr(parsed, "log_cache_events", False))

    # this will convert DbtConfigErrors into RuntimeExceptions
    # task could be any one of the task objects
    task = parsed.cls.from_args(args=parsed)

    # Set up logging
    log_path = None
    if task.config is not None:
        log_path = getattr(task.config, "log_path", None)
    log_manager.set_path(log_path)
    # if 'list' task: set stdout to WARN instead of INFO
    level_override = parsed.cls.pre_init_hook(parsed)
    setup_event_logger(log_path or "logs", level_override)

    fire_event(MainReportVersion(v=str(dbt.version.installed)))
    fire_event(MainReportArgs(args=args_to_dict(parsed)))

    if dbt.tracking.active_user is not None:  # mypy appeasement, always true
        fire_event(MainTrackingUserState(user_state=dbt.tracking.active_user.state()))

    results = None

    with track_run(task):
        results = task.run()
    return task, results


def _build_base_subparser():
    base_subparser = argparse.ArgumentParser(add_help=False)

    base_subparser.add_argument(
        "--project-dir",
        default=None,
        type=str,
        help="""
        Which directory to look in for the dbt_project.yml file.
        Default is the current working directory and its parents.
        """,
    )

    base_subparser.add_argument(
        "--profiles-dir",
        default=None,
        dest="sub_profiles_dir",  # Main cli arg precedes subcommand
        type=str,
        help="""
        Which directory to look in for the profiles.yml file. Default = {}
        """.format(
            DEFAULT_PROFILES_DIR
        ),
    )

    base_subparser.add_argument(
        "--profile",
        required=False,
        type=str,
        help="""
        Which profile to load. Overrides setting in dbt_project.yml.
        """,
    )

    base_subparser.add_argument(
        "-t",
        "--target",
        default=None,
        type=str,
        help="""
        Which target to load for the given profile
        """,
    )

    base_subparser.add_argument(
        "--vars",
        type=str,
        default="{}",
        help="""
        Supply variables to the project. This argument overrides variables
        defined in your dbt_project.yml file. This argument should be a YAML
        string, eg. '{my_variable: my_value}'
        """,
    )

    # if set, log all cache events. This is extremely verbose!
    base_subparser.add_argument(
        "--log-cache-events",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    base_subparser.set_defaults(defer=None, state=None)
    return base_subparser


def _build_docs_subparser(subparsers, base_subparser):
    docs_sub = subparsers.add_parser(
        "docs",
        help="""
        Generate or serve the documentation website for your project.
        """,
    )
    return docs_sub


def _build_source_subparser(subparsers, base_subparser):
    source_sub = subparsers.add_parser(
        "source",
        help="""
        Manage your project's sources
        """,
    )
    return source_sub


def _build_init_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "init",
        parents=[base_subparser],
        help="""
        Initialize a new DBT project.
        """,
    )
    sub.add_argument(
        "project_name",
        nargs="?",
        help="""
        Name of the new DBT project.
        """,
    )
    sub.add_argument(
        "-s",
        "--skip-profile-setup",
        dest="skip_profile_setup",
        action="store_true",
        help="""
        Skip interative profile setup.
        """,
    )
    sub.set_defaults(cls=init_task.InitTask, which="init", rpc_method=None)
    return sub


def _build_build_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "build",
        parents=[base_subparser],
        help="""
        Run all Seeds, Models, Snapshots, and tests in DAG order
        """,
    )
    sub.set_defaults(cls=build_task.BuildTask, which="build", rpc_method="build")
    sub.add_argument(
        "-x",
        "--fail-fast",
        dest="sub_fail_fast",
        action="store_true",
        help="""
        Stop execution upon a first failure.
        """,
    )
    sub.add_argument(
        "--store-failures",
        action="store_true",
        help="""
        Store test results (failing rows) in the database
        """,
    )
    sub.add_argument(
        "--indirect-selection",
        choices=["eager", "cautious"],
        default="eager",
        dest="indirect_selection",
        help="""
            Select all tests that are adjacent to selected resources,
            even if they those resources have been explicitly selected.
        """,
    )

    resource_values: List[str] = [str(s) for s in build_task.BuildTask.ALL_RESOURCE_VALUES] + [
        "all"
    ]
    sub.add_argument(
        "--resource-type",
        choices=resource_values,
        action="append",
        default=[],
        dest="resource_types",
    )
    # explicity don't support --models
    sub.add_argument(
        "-s",
        "--select",
        dest="select",
        nargs="+",
        help="""
            Specify the nodes to include.
        """,
    )
    _add_common_selector_arguments(sub)
    return sub


def _build_clean_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "clean",
        parents=[base_subparser],
        help="""
        Delete all folders in the clean-targets list
        (usually the dbt_packages and target directories.)
        """,
    )
    sub.set_defaults(cls=clean_task.CleanTask, which="clean", rpc_method=None)
    return sub


def _build_debug_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "debug",
        parents=[base_subparser],
        help="""
        Show some helpful information about dbt for debugging.

        Not to be confused with the --debug option which increases verbosity.
        """,
    )
    sub.add_argument(
        "--config-dir",
        action="store_true",
        help="""
        If specified, DBT will show path information for this project
        """,
    )
    _add_version_check(sub)
    sub.set_defaults(cls=debug_task.DebugTask, which="debug", rpc_method=None)
    return sub


def _build_deps_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "deps",
        parents=[base_subparser],
        help="""
        Pull the most recent version of the dependencies listed in packages.yml
        """,
    )
    sub.set_defaults(cls=deps_task.DepsTask, which="deps", rpc_method="deps")
    return sub


def _build_snapshot_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "snapshot",
        parents=[base_subparser],
        help="""
        Execute snapshots defined in your project
        """,
    )
    sub.add_argument(
        "--threads",
        type=int,
        required=False,
        help="""
        Specify number of threads to use while snapshotting tables.
        Overrides settings in profiles.yml.
        """,
    )
    sub.set_defaults(cls=snapshot_task.SnapshotTask, which="snapshot", rpc_method="snapshot")
    return sub


def _add_defer_argument(*subparsers):
    for sub in subparsers:
        sub.add_optional_argument_inverse(
            "--defer",
            enable_help="""
            If set, defer to the state variable for resolving unselected nodes.
            """,
            disable_help="""
            If set, do not defer to the state variable for resolving unselected
            nodes.
            """,
            default=flags.DEFER_MODE,
        )


def _build_run_subparser(subparsers, base_subparser):
    run_sub = subparsers.add_parser(
        "run",
        parents=[base_subparser],
        help="""
        Compile SQL and execute against the current target database.
        """,
    )
    run_sub.add_argument(
        "-x",
        "--fail-fast",
        dest="sub_fail_fast",
        action="store_true",
        help="""
        Stop execution upon a first failure.
        """,
    )

    run_sub.set_defaults(cls=run_task.RunTask, which="run", rpc_method="run")
    return run_sub


def _build_compile_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "compile",
        parents=[base_subparser],
        help="""
        Generates executable SQL from source, model, test, and analysis files.
        Compiled SQL files are written to the target/ directory.
        """,
    )
    sub.set_defaults(cls=compile_task.CompileTask, which="compile", rpc_method="compile")
    sub.add_argument("--parse-only", action="store_true")
    return sub


def _build_parse_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "parse",
        parents=[base_subparser],
        help="""
        Parses the project and provides information on performance
        """,
    )
    sub.set_defaults(cls=parse_task.ParseTask, which="parse", rpc_method="parse")
    sub.add_argument("--write-manifest", action="store_true")
    sub.add_argument("--compile", action="store_true")
    return sub


def _build_docs_generate_subparser(subparsers, base_subparser):
    # it might look like docs_sub is the correct parents entry, but that
    # will cause weird errors about 'conflicting option strings'.
    generate_sub = subparsers.add_parser("generate", parents=[base_subparser])
    generate_sub.set_defaults(
        cls=generate_task.GenerateTask, which="generate", rpc_method="docs.generate"
    )
    generate_sub.add_argument(
        "--no-compile",
        action="store_false",
        dest="compile",
        help="""
        Do not run "dbt compile" as part of docs generation
        """,
    )
    return generate_sub


def _add_common_selector_arguments(sub):
    sub.add_argument(
        "--exclude",
        required=False,
        nargs="+",
        help="""
            Specify the models to exclude.
        """,
    )
    sub.add_argument(
        "--selector",
        dest="selector_name",
        metavar="SELECTOR_NAME",
        help="""
        The selector name to use, as defined in selectors.yml
        """,
    )
    sub.add_argument(
        "--state",
        help="""
        If set, use the given directory as the source for json files to
        compare with this project.
        """,
        type=Path,
        default=flags.ARTIFACT_STATE_PATH,
    )


def _add_selection_arguments(*subparsers):
    for sub in subparsers:
        sub.add_argument(
            "-m",
            "--models",
            dest="select",
            nargs="+",
            help="""
                Specify the nodes to include.
            """,
        )
        sub.add_argument(
            "-s",
            "--select",
            dest="select",
            nargs="+",
            help="""
                Specify the nodes to include.
            """,
        )
        _add_common_selector_arguments(sub)


def _add_table_mutability_arguments(*subparsers):
    for sub in subparsers:
        sub.add_argument(
            "--full-refresh",
            action="store_true",
            help="""
            If specified, dbt will drop incremental models and
            fully-recalculate the incremental table from the model definition.
            """,
        )


def _add_version_check(sub):
    sub.add_argument(
        "--no-version-check",
        dest="sub_version_check",  # main cli arg precedes subcommands
        action="store_false",
        default=None,
        help="""
        If set, skip ensuring dbt's version matches the one specified in
        the dbt_project.yml file ('require-dbt-version')
        """,
    )


def _add_common_arguments(*subparsers):
    for sub in subparsers:
        sub.add_argument(
            "--threads",
            type=int,
            required=False,
            help="""
            Specify number of threads to use while executing models. Overrides
            settings in profiles.yml.
            """,
        )
        sub.add_argument(
            "--target-path",
            required=False,
            help="""
            Configure the 'target-path'. Only applies this setting for the
            current run. Overrides the 'DBT_TARGET_PATH' if it is set.
            """,
        )
        sub.add_argument(
            "--log-path",
            required=False,
            help="""
            Configure the 'log-path'. Only applies this setting for the
            current run. Overrides the 'DBT_LOG_PATH' if it is set.
            """,
        )
        _add_version_check(sub)


def _build_seed_subparser(subparsers, base_subparser):
    seed_sub = subparsers.add_parser(
        "seed",
        parents=[base_subparser],
        help="""
        Load data from csv files into your data warehouse.
        """,
    )
    seed_sub.add_argument(
        "--full-refresh",
        action="store_true",
        help="""
        Drop existing seed tables and recreate them
        """,
    )
    seed_sub.add_argument(
        "--show",
        action="store_true",
        help="""
        Show a sample of the loaded data in the terminal
        """,
    )
    seed_sub.set_defaults(cls=seed_task.SeedTask, which="seed", rpc_method="seed")
    return seed_sub


def _build_docs_serve_subparser(subparsers, base_subparser):
    serve_sub = subparsers.add_parser("serve", parents=[base_subparser])
    serve_sub.add_argument(
        "--port",
        default=8080,
        type=int,
        help="""
        Specify the port number for the docs server.
        """,
    )
    serve_sub.add_argument(
        "--no-browser",
        dest="open_browser",
        action="store_false",
    )
    serve_sub.set_defaults(cls=serve_task.ServeTask, which="serve", rpc_method=None)
    return serve_sub


def _build_test_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "test",
        parents=[base_subparser],
        help="""
        Runs tests on data in deployed models. Run this after `dbt run`
        """,
    )
    sub.add_argument(
        "-x",
        "--fail-fast",
        dest="sub_fail_fast",
        action="store_true",
        help="""
        Stop execution upon a first test failure.
        """,
    )
    sub.add_argument(
        "--store-failures",
        action="store_true",
        help="""
        Store test results (failing rows) in the database
        """,
    )
    sub.add_argument(
        "--indirect-selection",
        choices=["eager", "cautious"],
        default="eager",
        dest="indirect_selection",
        help="""
            Select all tests that are adjacent to selected resources,
            even if they those resources have been explicitly selected.
        """,
    )

    sub.set_defaults(cls=test_task.TestTask, which="test", rpc_method="test")
    return sub


def _build_source_freshness_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "freshness",
        parents=[base_subparser],
        help="""
        Snapshots the current freshness of the project's sources
        """,
        aliases=["snapshot-freshness"],
    )
    sub.add_argument(
        "-o",
        "--output",
        required=False,
        help="""
        Specify the output path for the json report. By default, outputs to
        target/sources.json
        """,
    )
    sub.add_argument(
        "--threads",
        type=int,
        required=False,
        help="""
        Specify number of threads to use. Overrides settings in profiles.yml
        """,
    )
    sub.set_defaults(
        cls=freshness_task.FreshnessTask,
        which="source-freshness",
        rpc_method="source-freshness",
    )
    sub.add_argument(
        "-s",
        "--select",
        dest="select",
        nargs="+",
        help="""
            Specify the nodes to include.
        """,
    )
    _add_common_selector_arguments(sub)
    return sub


def _build_list_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "list",
        parents=[base_subparser],
        help="""
        List the resources in your project
        """,
        aliases=["ls"],
    )
    sub.set_defaults(cls=list_task.ListTask, which="list", rpc_method=None)
    resource_values: List[str] = [str(s) for s in list_task.ListTask.ALL_RESOURCE_VALUES] + [
        "default",
        "all",
    ]
    sub.add_argument(
        "--resource-type",
        choices=resource_values,
        action="append",
        default=[],
        dest="resource_types",
    )
    sub.add_argument("--output", choices=["json", "name", "path", "selector"], default="selector")
    sub.add_argument("--output-keys")

    sub.add_argument(
        "-m",
        "--models",
        dest="models",
        nargs="+",
        help="""
        Specify the models to select and set the resource-type to 'model'.
        Mutually exclusive with '--select' (or '-s') and '--resource-type'
        """,
        metavar="SELECTOR",
        required=False,
    )
    sub.add_argument(
        "-s",
        "--select",
        dest="select",
        nargs="+",
        help="""
            Specify the nodes to include.
        """,
        metavar="SELECTOR",
        required=False,
    )
    sub.add_argument(
        "--indirect-selection",
        choices=["eager", "cautious"],
        default="eager",
        dest="indirect_selection",
        help="""
            Select all tests that are adjacent to selected resources,
            even if they those resources have been explicitly selected.
        """,
    )
    _add_common_selector_arguments(sub)

    return sub


def _build_run_operation_subparser(subparsers, base_subparser):
    sub = subparsers.add_parser(
        "run-operation",
        parents=[base_subparser],
        help="""
        Run the named macro with any supplied arguments.
        """,
    )
    sub.add_argument(
        "macro",
        help="""
        Specify the macro to invoke. dbt will call this macro with the supplied
        arguments and then exit
        """,
    )
    sub.add_argument(
        "--args",
        type=str,
        default="{}",
        help="""
        Supply arguments to the macro. This dictionary will be mapped to the
        keyword arguments defined in the selected macro. This argument should
        be a YAML string, eg. '{my_variable: my_value}'
        """,
    )
    sub.set_defaults(
        cls=run_operation_task.RunOperationTask, which="run-operation", rpc_method="run-operation"
    )
    return sub


def parse_args(args, cls=DBTArgumentParser):
    p = cls(
        prog="dbt",
        description="""
        An ELT tool for managing your SQL transformations and data models.
        For more documentation on these commands, visit: docs.getdbt.com
        """,
        epilog="""
        Specify one of these sub-commands and you can find more help from
        there.
        """,
    )

    p.add_argument(
        "--version",
        action="dbtversion",
        help="""
        Show version information
        """,
    )

    p.add_argument(
        "-r",
        "--record-timing-info",
        default=None,
        type=str,
        help="""
        When this option is passed, dbt will output low-level timing stats to
        the specified file. Example: `--record-timing-info output.profile`
        """,
    )

    p.add_argument(
        "-d",
        "--debug",
        action="store_true",
        default=None,
        help="""
        Display debug logging during dbt execution. Useful for debugging and
        making bug reports.
        """,
    )

    p.add_argument(
        "--log-format",
        choices=["text", "json", "default"],
        default=None,
        help="""Specify the log format, overriding the command's default.""",
    )

    p.add_argument(
        "--no-write-json",
        action="store_false",
        default=None,
        dest="write_json",
        help="""
        If set, skip writing the manifest and run_results.json files to disk
        """,
    )
    colors_flag = p.add_mutually_exclusive_group()
    colors_flag.add_argument(
        "--use-colors",
        action="store_const",
        const=True,
        default=None,
        dest="use_colors",
        help="""
        Colorize the output DBT prints to the terminal. Output is colorized by
        default and may also be set in a profile or at the command line.
        Mutually exclusive with --no-use-colors
        """,
    )
    colors_flag.add_argument(
        "--no-use-colors",
        action="store_const",
        const=False,
        dest="use_colors",
        help="""
        Do not colorize the output DBT prints to the terminal. Output is
        colorized by default and may also be set in a profile or at the
        command line.
        Mutually exclusive with --use-colors
        """,
    )

    p.add_argument(
        "--printer-width",
        dest="printer_width",
        help="""
        Sets the width of terminal output
        """,
    )

    p.add_argument(
        "--warn-error",
        action="store_true",
        default=None,
        help="""
        If dbt would normally warn, instead raise an exception. Examples
        include --models that selects nothing, deprecations, configurations
        with no associated models, invalid test configurations, and missing
        sources/refs in tests.
        """,
    )

    p.add_argument(
        "--no-version-check",
        dest="version_check",
        action="store_false",
        default=None,
        help="""
        If set, skip ensuring dbt's version matches the one specified in
        the dbt_project.yml file ('require-dbt-version')
        """,
    )

    p.add_optional_argument_inverse(
        "--partial-parse",
        enable_help="""
        Allow for partial parsing by looking for and writing to a pickle file
        in the target directory. This overrides the user configuration file.
        """,
        disable_help="""
        Disallow partial parsing. This overrides the user configuration file.
        """,
    )

    # if set, run dbt in single-threaded mode: thread count is ignored, and
    # calls go through `map` instead of the thread pool. This is useful for
    # getting performance information about aspects of dbt that normally run in
    # a thread, as the profiler ignores child threads. Users should really
    # never use this.
    p.add_argument(
        "--single-threaded",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    # if set, will use the latest features from the static parser instead of
    # the stable static parser.
    p.add_argument(
        "--use-experimental-parser",
        action="store_true",
        default=None,
        help="""
        Enables experimental parsing features.
        """,
    )

    # if set, will disable the use of the stable static parser and instead
    # always rely on jinja rendering.
    p.add_argument(
        "--no-static-parser",
        default=None,
        dest="static_parser",
        action="store_false",
        help="""
        Disables the static parser.
        """,
    )

    p.add_argument(
        "--profiles-dir",
        default=None,
        dest="profiles_dir",
        type=str,
        help="""
        Which directory to look in for the profiles.yml file. Default = {}
        """.format(
            DEFAULT_PROFILES_DIR
        ),
    )

    p.add_argument(
        "--no-anonymous-usage-stats",
        action="store_false",
        default=None,
        dest="send_anonymous_usage_stats",
        help="""
        Do not send anonymous usage stat to dbt Labs
        """,
    )

    p.add_argument(
        "-x",
        "--fail-fast",
        dest="fail_fast",
        action="store_true",
        default=None,
        help="""
        Stop execution upon a first failure.
        """,
    )

    p.add_argument(
        "--event-buffer-size",
        dest="event_buffer_size",
        help="""
        Sets the max number of events to buffer in EVENT_HISTORY
        """,
    )

    p.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        default=None,
        help="""
        Suppress all non-error logging to stdout. Does not affect
        {{ print() }} macro calls.
        """,
    )

    p.add_argument(
        "--no-print",
        action="store_true",
        default=None,
        help="""
        Suppress all {{ print() }} macro calls.
        """,
    )

    schema_cache_flag = p.add_mutually_exclusive_group()
    schema_cache_flag.add_argument(
        "--cache-selected-only",
        action="store_const",
        const=True,
        default=None,
        dest="cache_selected_only",
        help="""
        Pre cache database objects relevant to selected resource only.
        """,
    )
    schema_cache_flag.add_argument(
        "--no-cache-selected-only",
        action="store_const",
        const=False,
        dest="cache_selected_only",
        help="""
        Pre cache all database objects related to the project.
        """,
    )

    subs = p.add_subparsers(title="Available sub-commands")

    base_subparser = _build_base_subparser()

    # make the subcommands that have their own subcommands
    docs_sub = _build_docs_subparser(subs, base_subparser)
    docs_subs = docs_sub.add_subparsers(title="Available sub-commands")
    source_sub = _build_source_subparser(subs, base_subparser)
    source_subs = source_sub.add_subparsers(title="Available sub-commands")

    _build_init_subparser(subs, base_subparser)
    _build_clean_subparser(subs, base_subparser)
    _build_debug_subparser(subs, base_subparser)
    _build_deps_subparser(subs, base_subparser)
    _build_list_subparser(subs, base_subparser)

    build_sub = _build_build_subparser(subs, base_subparser)
    snapshot_sub = _build_snapshot_subparser(subs, base_subparser)
    run_sub = _build_run_subparser(subs, base_subparser)
    compile_sub = _build_compile_subparser(subs, base_subparser)
    parse_sub = _build_parse_subparser(subs, base_subparser)
    generate_sub = _build_docs_generate_subparser(docs_subs, base_subparser)
    test_sub = _build_test_subparser(subs, base_subparser)
    seed_sub = _build_seed_subparser(subs, base_subparser)
    # --threads, --no-version-check
    _add_common_arguments(
        run_sub, compile_sub, generate_sub, test_sub, seed_sub, parse_sub, build_sub
    )
    # --select, --exclude
    # list_sub sets up its own arguments.
    _add_selection_arguments(run_sub, compile_sub, generate_sub, test_sub, snapshot_sub, seed_sub)
    # --defer
    _add_defer_argument(run_sub, test_sub, build_sub, snapshot_sub)
    # --full-refresh
    _add_table_mutability_arguments(run_sub, compile_sub, build_sub)

    _build_docs_serve_subparser(docs_subs, base_subparser)
    _build_source_freshness_subparser(source_subs, base_subparser)
    _build_run_operation_subparser(subs, base_subparser)

    if len(args) == 0:
        p.print_help()
        sys.exit(1)

    parsed = p.parse_args(args)

    # profiles_dir is set before subcommands and after, so normalize
    if hasattr(parsed, "sub_profiles_dir"):
        if parsed.sub_profiles_dir is not None:
            parsed.profiles_dir = parsed.sub_profiles_dir
        delattr(parsed, "sub_profiles_dir")
    if hasattr(parsed, "profiles_dir"):
        if parsed.profiles_dir is None:
            parsed.profiles_dir = flags.PROFILES_DIR
        else:
            parsed.profiles_dir = os.path.abspath(parsed.profiles_dir)
            # needs to be set before the other flags, because it's needed to
            # read the profile that contains them
            flags.PROFILES_DIR = parsed.profiles_dir

    # version_check is set before subcommands and after, so normalize
    if hasattr(parsed, "sub_version_check"):
        if parsed.sub_version_check is False:
            parsed.version_check = False
        delattr(parsed, "sub_version_check")

    # fail_fast is set before subcommands and after, so normalize
    if hasattr(parsed, "sub_fail_fast"):
        if parsed.sub_fail_fast is True:
            parsed.fail_fast = True
        delattr(parsed, "sub_fail_fast")

    if getattr(parsed, "project_dir", None) is not None:
        expanded_user = os.path.expanduser(parsed.project_dir)
        parsed.project_dir = os.path.abspath(expanded_user)

    if not hasattr(parsed, "which"):
        # the user did not provide a valid subcommand. trigger the help message
        # and exit with a error
        p.print_help()
        p.exit(1)

    return parsed



import os
import functools
from typing import List

from dbt import semver
from dbt.clients import registry, system
from dbt.contracts.project import (
    RegistryPackageMetadata,
    RegistryPackage,
)
from dbt.deps.base import PinnedPackage, UnpinnedPackage, get_downloads_path
from dbt.exceptions import (
    package_version_not_found,
    VersionsNotCompatibleException,
    DependencyException,
    package_not_found,
)
from dbt.utils import _connection_exception_retry as connection_exception_retry


class RegistryPackageMixin:
    def __init__(self, package: str) -> None:
        super().__init__()
        self.package = package

    @property
    def name(self):
        return self.package

    def source_type(self) -> str:
        return "hub"


class RegistryPinnedPackage(RegistryPackageMixin, PinnedPackage):
    def __init__(self, package: str, version: str, version_latest: str) -> None:
        super().__init__(package)
        self.version = version
        self.version_latest = version_latest

    @property
    def name(self):
        return self.package

    def source_type(self):
        return "hub"

    def get_version(self):
        return self.version

    def get_version_latest(self):
        return self.version_latest

    def nice_version_name(self):
        return "version {}".format(self.version)

    def _fetch_metadata(self, project, renderer) -> RegistryPackageMetadata:
        dct = registry.package_version(self.package, self.version)
        return RegistryPackageMetadata.from_dict(dct)

    def install(self, project, renderer):
        metadata = self.fetch_metadata(project, renderer)

        tar_name = "{}.{}.tar.gz".format(self.package, self.version)
        tar_path = os.path.realpath(os.path.join(get_downloads_path(), tar_name))
        system.make_directory(os.path.dirname(tar_path))

        download_url = metadata.downloads.tarball
        deps_path = project.packages_install_path
        package_name = self.get_project_name(project, renderer)

        download_untar_fn = functools.partial(
            self.download_and_untar, download_url, tar_path, deps_path, package_name
        )
        connection_exception_retry(download_untar_fn, 5)

    def download_and_untar(self, download_url, tar_path, deps_path, package_name):
        """
        Sometimes the download of the files fails and we want to retry.  Sometimes the
        download appears successful but the file did not make it through as expected
        (generally due to a github incident).  Either way we want to retry downloading
        and untarring to see if we can get a success.  Call this within
        `_connection_exception_retry`
        """

        system.download(download_url, tar_path)
        system.untar_package(tar_path, deps_path, package_name)


class RegistryUnpinnedPackage(RegistryPackageMixin, UnpinnedPackage[RegistryPinnedPackage]):
    def __init__(
        self, package: str, versions: List[semver.VersionSpecifier], install_prerelease: bool
    ) -> None:
        super().__init__(package)
        self.versions = versions
        self.install_prerelease = install_prerelease

    def _check_in_index(self):
        index = registry.index_cached()
        if self.package not in index:
            package_not_found(self.package)

    @classmethod
    def from_contract(cls, contract: RegistryPackage) -> "RegistryUnpinnedPackage":
        raw_version = contract.get_versions()

        versions = [semver.VersionSpecifier.from_version_string(v) for v in raw_version]
        return cls(
            package=contract.package,
            versions=versions,
            install_prerelease=bool(contract.install_prerelease),
        )

    def incorporate(self, other: "RegistryUnpinnedPackage") -> "RegistryUnpinnedPackage":
        return RegistryUnpinnedPackage(
            package=self.package,
            install_prerelease=self.install_prerelease,
            versions=self.versions + other.versions,
        )

    def resolved(self) -> RegistryPinnedPackage:
        self._check_in_index()
        try:
            range_ = semver.reduce_versions(*self.versions)
        except VersionsNotCompatibleException as e:
            new_msg = "Version error for package {}: {}".format(self.name, e)
            raise DependencyException(new_msg) from e

        available = registry.get_available_versions(self.package)
        prerelease_version_specified = any(bool(version.prerelease) for version in self.versions)
        installable = semver.filter_installable(
            available, self.install_prerelease or prerelease_version_specified
        )
        available_latest = installable[-1]

        # for now, pick a version and then recurse. later on,
        # we'll probably want to traverse multiple options
        # so we can match packages. not going to make a difference
        # right now.
        target = semver.resolve_to_specific_version(range_, installable)
        if not target:
            package_version_not_found(self.package, range_, installable)
        return RegistryPinnedPackage(
            package=self.package, version=target, version_latest=available_latest
        )



import os
import threading
import time
from abc import ABCMeta, abstractmethod
from typing import Type, Union, Dict, Any, Optional

from dbt import tracking
from dbt import flags
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.results import (
    NodeStatus,
    RunResult,
    collect_timing_info,
    RunStatus,
    RunningStatus,
)
from dbt.exceptions import (
    NotImplementedException,
    CompilationException,
    RuntimeException,
    InternalException,
)
from dbt.logger import log_manager
import dbt.events.functions as event_logger
from dbt.events.functions import fire_event
from dbt.events.types import (
    DbtProjectError,
    DbtProjectErrorException,
    DbtProfileError,
    DbtProfileErrorException,
    ProfileListTitle,
    ListSingleProfile,
    NoDefinedProfiles,
    ProfileHelpMessage,
    CatchableExceptionOnRun,
    InternalExceptionOnRun,
    GenericExceptionOnRun,
    NodeConnectionReleaseError,
    PrintDebugStackTrace,
    SkippingDetails,
    PrintSkipBecauseError,
    NodeCompiling,
    NodeExecuting,
)
from .printer import print_run_result_error

from dbt.adapters.factory import register_adapter
from dbt.config import RuntimeConfig, Project
from dbt.config.profile import read_profile
import dbt.exceptions


class NoneConfig:
    @classmethod
    def from_args(cls, args):
        return None


def read_profiles(profiles_dir=None):
    """This is only used for some error handling"""
    if profiles_dir is None:
        profiles_dir = flags.PROFILES_DIR

    raw_profiles = read_profile(profiles_dir)

    if raw_profiles is None:
        profiles = {}
    else:
        profiles = {k: v for (k, v) in raw_profiles.items() if k != "config"}

    return profiles


class BaseTask(metaclass=ABCMeta):
    ConfigType: Union[Type[NoneConfig], Type[Project]] = NoneConfig

    def __init__(self, args, config):
        self.args = args
        self.args.single_threaded = False
        self.config = config

    @classmethod
    def pre_init_hook(cls, args):
        """A hook called before the task is initialized."""
        if args.log_format == "json":
            log_manager.format_json()
            # we're mutating the initialized, but not-yet-configured event logger
            # because it's being configured too late -- bad! TODO refactor!
            event_logger.format_json = True
        else:
            log_manager.format_text()

    @classmethod
    def set_log_format(cls):
        if flags.LOG_FORMAT == "json":
            log_manager.format_json()
            # we're mutating the initialized, but not-yet-configured event logger
            # because it's being configured too late -- bad! TODO refactor!
            event_logger.format_json = True
        else:
            log_manager.format_text()

    @classmethod
    def from_args(cls, args):
        try:
            # This is usually RuntimeConfig but will be UnsetProfileConfig
            # for the clean or deps tasks
            config = cls.ConfigType.from_args(args)
        except dbt.exceptions.DbtProjectError as exc:
            fire_event(DbtProjectError())
            fire_event(DbtProjectErrorException(exc=exc))

            tracking.track_invalid_invocation(args=args, result_type=exc.result_type)
            raise dbt.exceptions.RuntimeException("Could not run dbt") from exc
        except dbt.exceptions.DbtProfileError as exc:
            fire_event(DbtProfileError())
            fire_event(DbtProfileErrorException(exc=exc))

            all_profiles = read_profiles(flags.PROFILES_DIR).keys()

            if len(all_profiles) > 0:
                fire_event(ProfileListTitle())
                for profile in all_profiles:
                    fire_event(ListSingleProfile(profile=profile))
            else:
                fire_event(NoDefinedProfiles())

            fire_event(ProfileHelpMessage())

            tracking.track_invalid_invocation(args=args, result_type=exc.result_type)
            raise dbt.exceptions.RuntimeException("Could not run dbt") from exc
        return cls(args, config)

    @abstractmethod
    def run(self):
        raise dbt.exceptions.NotImplementedException("Not Implemented")

    def interpret_results(self, results):
        return True


def get_nearest_project_dir(args):
    # If the user provides an explicit project directory, use that
    # but don't look at parent directories.
    if args.project_dir:
        project_file = os.path.join(args.project_dir, "dbt_project.yml")
        if os.path.exists(project_file):
            return args.project_dir
        else:
            raise dbt.exceptions.RuntimeException(
                "fatal: Invalid --project-dir flag. Not a dbt project. "
                "Missing dbt_project.yml file"
            )

    root_path = os.path.abspath(os.sep)
    cwd = os.getcwd()

    while cwd != root_path:
        project_file = os.path.join(cwd, "dbt_project.yml")
        if os.path.exists(project_file):
            return cwd
        cwd = os.path.dirname(cwd)

    raise dbt.exceptions.RuntimeException(
        "fatal: Not a dbt project (or any of the parent directories). "
        "Missing dbt_project.yml file"
    )


def move_to_nearest_project_dir(args):
    nearest_project_dir = get_nearest_project_dir(args)
    os.chdir(nearest_project_dir)


class ConfiguredTask(BaseTask):
    ConfigType = RuntimeConfig

    def __init__(self, args, config):
        super().__init__(args, config)
        register_adapter(self.config)

    @classmethod
    def from_args(cls, args):
        move_to_nearest_project_dir(args)
        return super().from_args(args)


class ExecutionContext:
    """During execution and error handling, dbt makes use of mutable state:
    timing information and the newest (compiled vs executed) form of the node.
    """

    def __init__(self, node):
        self.timing = []
        self.node = node


class BaseRunner(metaclass=ABCMeta):
    def __init__(self, config, adapter, node, node_index, num_nodes):
        self.config = config
        self.adapter = adapter
        self.node = node
        self.node_index = node_index
        self.num_nodes = num_nodes

        self.skip = False
        self.skip_cause: Optional[RunResult] = None

    @abstractmethod
    def compile(self, manifest: Manifest) -> Any:
        pass

    def get_result_status(self, result) -> Dict[str, str]:
        if result.status == NodeStatus.Error:
            return {"node_status": "error", "node_error": str(result.message)}
        elif result.status == NodeStatus.Skipped:
            return {"node_status": "skipped"}
        elif result.status == NodeStatus.Fail:
            return {"node_status": "failed"}
        elif result.status == NodeStatus.Warn:
            return {"node_status": "warn"}
        else:
            return {"node_status": "passed"}

    def run_with_hooks(self, manifest):
        if self.skip:
            return self.on_skip()

        # no before/after printing for ephemeral mdoels
        if not self.node.is_ephemeral_model:
            self.before_execute()

        result = self.safe_run(manifest)

        if not self.node.is_ephemeral_model:
            self.after_execute(result)

        return result

    def _build_run_result(
        self,
        node,
        start_time,
        status,
        timing_info,
        message,
        agate_table=None,
        adapter_response=None,
        failures=None,
    ):
        execution_time = time.time() - start_time
        thread_id = threading.current_thread().name
        if adapter_response is None:
            adapter_response = {}
        return RunResult(
            status=status,
            thread_id=thread_id,
            execution_time=execution_time,
            timing=timing_info,
            message=message,
            node=node,
            agate_table=agate_table,
            adapter_response=adapter_response,
            failures=failures,
        )

    def error_result(self, node, message, start_time, timing_info):
        return self._build_run_result(
            node=node,
            start_time=start_time,
            status=RunStatus.Error,
            timing_info=timing_info,
            message=message,
        )

    def ephemeral_result(self, node, start_time, timing_info):
        return self._build_run_result(
            node=node,
            start_time=start_time,
            status=RunStatus.Success,
            timing_info=timing_info,
            message=None,
        )

    def from_run_result(self, result, start_time, timing_info):
        return self._build_run_result(
            node=result.node,
            start_time=start_time,
            status=result.status,
            timing_info=timing_info,
            message=result.message,
            agate_table=result.agate_table,
            adapter_response=result.adapter_response,
            failures=result.failures,
        )

    def skip_result(self, node, message):
        thread_id = threading.current_thread().name
        return RunResult(
            status=RunStatus.Skipped,
            thread_id=thread_id,
            execution_time=0,
            timing=[],
            message=message,
            node=node,
            adapter_response={},
            failures=None,
        )

    def compile_and_execute(self, manifest, ctx):
        result = None
        with self.adapter.connection_for(self.node):
            ctx.node._event_status["node_status"] = RunningStatus.Compiling
            fire_event(
                NodeCompiling(
                    node_info=ctx.node.node_info,
                    unique_id=ctx.node.unique_id,
                )
            )
            with collect_timing_info("compile") as timing_info:
                # if we fail here, we still have a compiled node to return
                # this has the benefit of showing a build path for the errant
                # model
                ctx.node = self.compile(manifest)
            ctx.timing.append(timing_info)

            # for ephemeral nodes, we only want to compile, not run
            if not ctx.node.is_ephemeral_model:
                ctx.node._event_status["node_status"] = RunningStatus.Executing
                fire_event(
                    NodeExecuting(
                        node_info=ctx.node.node_info,
                        unique_id=ctx.node.unique_id,
                    )
                )
                with collect_timing_info("execute") as timing_info:
                    result = self.run(ctx.node, manifest)
                    ctx.node = result.node

                ctx.timing.append(timing_info)

        return result

    def _handle_catchable_exception(self, e, ctx):
        if e.node is None:
            e.add_node(ctx.node)

        fire_event(CatchableExceptionOnRun(exc=e))
        return str(e)

    def _handle_internal_exception(self, e, ctx):
        fire_event(InternalExceptionOnRun(build_path=self.node.build_path, exc=e))
        return str(e)

    def _handle_generic_exception(self, e, ctx):
        fire_event(
            GenericExceptionOnRun(
                build_path=self.node.build_path,
                unique_id=self.node.unique_id,
                exc=str(e),  # TODO: unstring this when serialization is fixed
            )
        )
        fire_event(PrintDebugStackTrace())

        return str(e)

    def handle_exception(self, e, ctx):
        catchable_errors = (CompilationException, RuntimeException)
        if isinstance(e, catchable_errors):
            error = self._handle_catchable_exception(e, ctx)
        elif isinstance(e, InternalException):
            error = self._handle_internal_exception(e, ctx)
        else:
            error = self._handle_generic_exception(e, ctx)
        return error

    def safe_run(self, manifest):
        started = time.time()
        ctx = ExecutionContext(self.node)
        error = None
        result = None

        try:
            result = self.compile_and_execute(manifest, ctx)
        except Exception as e:
            error = self.handle_exception(e, ctx)
        finally:
            exc_str = self._safe_release_connection()

            # if releasing failed and the result doesn't have an error yet, set
            # an error
            if (
                exc_str is not None
                and result is not None
                and result.status != NodeStatus.Error
                and error is None
            ):
                error = exc_str

        if error is not None:
            # we could include compile time for runtime errors here
            result = self.error_result(ctx.node, error, started, [])
        elif result is not None:
            result = self.from_run_result(result, started, ctx.timing)
        else:
            result = self.ephemeral_result(ctx.node, started, ctx.timing)
        return result

    def _safe_release_connection(self):
        """Try to release a connection. If an exception is hit, log and return
        the error string.
        """
        try:
            self.adapter.release_connection()
        except Exception as exc:
            fire_event(NodeConnectionReleaseError(node_name=self.node.name, exc=exc))
            return str(exc)

        return None

    def before_execute(self):
        raise NotImplementedException()

    def execute(self, compiled_node, manifest):
        raise NotImplementedException()

    def run(self, compiled_node, manifest):
        return self.execute(compiled_node, manifest)

    def after_execute(self, result):
        raise NotImplementedException()

    def _skip_caused_by_ephemeral_failure(self):
        if self.skip_cause is None or self.skip_cause.node is None:
            return False
        return self.skip_cause.node.is_ephemeral_model

    def on_skip(self):
        schema_name = self.node.schema
        node_name = self.node.name

        error_message = None
        if not self.node.is_ephemeral_model:
            # if this model was skipped due to an upstream ephemeral model
            # failure, print a special 'error skip' message.
            if self._skip_caused_by_ephemeral_failure():
                fire_event(
                    PrintSkipBecauseError(
                        schema=schema_name,
                        relation=node_name,
                        index=self.node_index,
                        total=self.num_nodes,
                    )
                )
                print_run_result_error(result=self.skip_cause, newline=False)
                if self.skip_cause is None:  # mypy appeasement
                    raise InternalException(
                        "Skip cause not set but skip was somehow caused by " "an ephemeral failure"
                    )
                # set an error so dbt will exit with an error code
                error_message = (
                    "Compilation Error in {}, caused by compilation error "
                    "in referenced ephemeral model {}".format(
                        self.node.unique_id, self.skip_cause.node.unique_id
                    )
                )
            else:
                fire_event(
                    SkippingDetails(
                        resource_type=self.node.resource_type,
                        schema=schema_name,
                        node_name=node_name,
                        index=self.node_index,
                        total=self.num_nodes,
                        node_info=self.node.node_info,
                    )
                )

        node_result = self.skip_result(self.node, error_message)
        return node_result

    def do_skip(self, cause=None):
        self.skip = True
        self.skip_cause = cause
